# Copyright 2021 DMTF. All rights reserved.
# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import math

from oslo_log import log
from oslo_utils import units
import sushy

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import raid
from ironic.common import states
from ironic.conductor import periodics
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.redfish import utils as redfish_utils

LOG = log.getLogger(__name__)
METRICS = metrics_utils.get_metrics_logger(__name__)


# TODO(billdodd): double-check all these values
RAID_LEVELS = {
    '0': {
        'min_disks': 1,
        'max_disks': 1000,
        'type': 'simple',
        'raid_type': 'RAID0',
        'overhead': 0
    },
    '1': {
        'min_disks': 2,
        'max_disks': 2,
        'type': 'simple',
        'raid_type': 'RAID1',
        'overhead': 1
    },
    '5': {
        'min_disks': 3,
        'max_disks': 1000,
        'type': 'simple',
        'raid_type': 'RAID5',
        'overhead': 1
    },
    '6': {
        'min_disks': 4,
        'max_disks': 1000,
        'type': 'simple',
        'raid_type': 'RAID6',
        'overhead': 2
    },
    '1+0': {
        'type': 'spanned',
        'raid_type': 'RAID10',
        'span_type': '1'
    },
    '5+0': {
        'type': 'spanned',
        'raid_type': 'RAID50',
        'span_type': '5'
    },
    '6+0': {
        'type': 'spanned',
        'raid_type': 'RAID60',
        'span_type': '6'
    }
}

PROTOCOL_MAP = {
    sushy.PROTOCOL_TYPE_SAS: raid.SAS,
    sushy.PROTOCOL_TYPE_SATA: raid.SATA
}


def convert_drive_units(logical_disks, node):
    """Convert size in logical_disks from gb to bytes"""
    for disk in logical_disks:
        if disk['size_gb'] == 'MAX' and 'physical_disks' not in disk:
            raise exception.InvalidParameterValue(
                _("create_configuration called with invalid "
                  "target_raid_configuration for node %(node_id)s. "
                  "'physical_disks' is missing from logical_disk while "
                  "'size_gb'='MAX' was requested: "
                  "%(logical_disk)s") % {'node_id': node.uuid,
                                         'logical_disk': disk})

        if disk['size_gb'] == 'MAX':
            disk['size_bytes'] = 'MAX'
        else:
            disk['size_bytes'] = disk['size_gb'] * units.Gi

        del disk['size_gb']


def get_physical_disks(node):
    """Get the physical drives of the node for RAID controllers.

    :param node: an ironic node object.
    :returns: a list of Drive objects from sushy
    :raises: RedfishConnectionError when it fails to connect to Redfish
    :raises: RedfishError if there is an error getting the drives via Redfish
    """
    system = redfish_utils.get_system(node)

    disks = []
    disk_to_storage = {}
    try:
        collection = system.storage
        for storage in collection.get_members():
            controller = redfish_utils.get_first_controller(storage)
            if controller and controller.raid_types == []:
                continue
            disks.extend(storage.drives)
            for drive in storage.drives:
                disk_to_storage[drive] = storage
    except sushy.exceptions.SushyError as exc:
        error_msg = _('Cannot get the list of physical disks for node '
                      '%(node_uuid)s. Reason: %(error)s.' %
                      {'node_uuid': node.uuid, 'error': exc})
        LOG.error(error_msg)
        raise exception.RedfishError(error=exc)
    return disks, disk_to_storage


def _raise_raid_level_not_supported(raid_level):
    """Helper function for the 'RAID level is not supported' error

    :param raid_level: RAID level of the virtual disk
    :raises: exception.RedfishError
    """
    reason = (_('RAID level %(raid_level)s is not supported by the '
                'driver. Supported RAID levels: %(supported_raid_levels)s')
              % {'raid_level': raid_level,
                 'supported_raid_levels': ', '.join(RAID_LEVELS)})
    raise exception.RedfishError(error=reason)


def _raid_level_overhead(raid_level, spans_count=1):
    """Calculate the drive overhead for the given RAID level

    Drive overhead is the number of additional drives required to hold the
    the redundant data needed for mirrored volumes and the parity checksums
    for volumes with parity.

    :param raid_level: RAID level of the virtual disk
    :param spans_count: number of spans for the virtual disk
    :return: the number of drives of overhead
    :raises: RedfishError if RAID level is not supported
    """
    try:
        raid_level_info = RAID_LEVELS[raid_level]
    except KeyError:
        _raise_raid_level_not_supported(raid_level)

    if raid_level_info['type'] == 'spanned':
        if spans_count <= 1:
            reason = _('Spanned RAID volumes cannot contain a single span')
            raise exception.RedfishError(error=reason)

        span_type = raid_level_info['span_type']
        raid_level_info = RAID_LEVELS[span_type]

    return raid_level_info['overhead'] * spans_count


def _max_volume_size_bytes(raid_level, physical_disks, free_space_bytes,
                           spans_count=1, stripe_size_kb=64 * units.Ki):
    # restrict the size to the smallest available space
    free_spaces = [free_space_bytes[disk] for disk in physical_disks]
    size_kb = min(free_spaces) // units.Ki

    # NOTE(ifarkas): using math.floor so we get a volume size that does not
    #                exceed the available space
    stripes_per_disk = int(math.floor(float(size_kb) / stripe_size_kb))

    disks_count = len(physical_disks)
    overhead_disks_count = _raid_level_overhead(raid_level, spans_count)
    if disks_count <= overhead_disks_count:
        reason = _('The number of physical drives (%(drives)s) is too few for '
                   'the required number of overhead drives (%(overhead)s)' %
                   {'drives': disks_count, 'overhead': overhead_disks_count})
        raise exception.RedfishError(error=reason)

    max_volume_size_bytes = int(
        stripes_per_disk * stripe_size_kb
        * (disks_count - overhead_disks_count) * units.Ki)
    return max_volume_size_bytes


def _volume_usage_per_disk_bytes(logical_disk, physical_disks, spans_count=1,
                                 stripe_size_kb=64 * units.Ki):
    disks_count = len(physical_disks)
    overhead_disks_count = _raid_level_overhead(logical_disk['raid_level'],
                                                spans_count)
    volume_size_kb = logical_disk['size_bytes'] // units.Ki
    # NOTE(ifarkas): using math.ceil so we get the largest disk usage
    #                possible, so we can avoid over-committing
    stripes_per_volume = math.ceil(float(volume_size_kb) / stripe_size_kb)

    stripes_per_disk = math.ceil(
        float(stripes_per_volume) / (disks_count - overhead_disks_count))
    volume_usage_per_disk_bytes = int(
        stripes_per_disk * stripe_size_kb * units.Ki)
    return volume_usage_per_disk_bytes


def _calculate_spans(raid_level, disks_count):
    """Calculates number of spans for a RAID level given a physical disk count

    :param raid_level: RAID level of the virtual disk.
    :param disks_count: number of physical disks used for the virtual disk.
    :returns: number of spans.
    """
    if raid_level in ['0', '1', '5', '6']:
        return 1
    elif raid_level in ['5+0', '6+0']:
        return 2
    elif raid_level in ['1+0']:
        return disks_count >> 1
    else:
        reason = (_('Cannot calculate spans for RAID level "%s"') %
                  raid_level)
        raise exception.RedfishError(error=reason)


def _calculate_volume_props(logical_disk, physical_disks, free_space_bytes,
                            disk_to_storage):
    """Calculate specific properties of the volume and update logical_disk dict

    Calculates various properties like span_depth and span_length for the
    logical disk to be created. Converts the size_gb property to size_bytes for
    use by sushy. Also performs checks to be sure the amount of physical space
    required for the logical disk is available.

    :param logical_disk: properties of the logical disk to create as
           specified by the operator.
    :param physical_disks: list of drives available on the node.
    :param free_space_bytes: dict mapping drives to their available space.
    :param disk_to_storage: dict mapping drives to their storage.
    :raises: RedfishError if physical drives cannot fulfill the logical disk.
    """
    # TODO(billdodd): match e.g. {'size': '> 100'} -> oslo_utils.specs_matcher
    selected_disks = [disk for disk in physical_disks
                      if disk.identity in logical_disk['physical_disks']]

    spans_count = _calculate_spans(
        logical_disk['raid_level'], len(selected_disks))

    if spans_count == 0 or len(selected_disks) % spans_count != 0:
        error_msg = _('For RAID level %(raid_level)s, the number of physical '
                      'disks provided (%(num_disks)s) must be a multiple of '
                      'the spans count (%(spans_count)s)' %
                      {'raid_level': logical_disk['raid_level'],
                       'num_disks': len(selected_disks),
                       'spans_count': spans_count})
        raise exception.RedfishError(error=error_msg)

    disks_per_span = len(selected_disks) / spans_count

    # TODO(billdodd): confirm this?
    # Best practice is to not pass span_length and span_depth when creating a
    # RAID10. Redfish will dynamically calculate these values using maximum
    # values obtained from the RAID controller.
    logical_disk['span_depth'] = None
    logical_disk['span_length'] = None
    if logical_disk['raid_level'] != '1+0':
        logical_disk['span_depth'] = spans_count
        logical_disk['span_length'] = disks_per_span

    max_volume_size_bytes = _max_volume_size_bytes(
        logical_disk['raid_level'], selected_disks, free_space_bytes,
        spans_count=spans_count)

    if logical_disk['size_bytes'] == 'MAX':
        if max_volume_size_bytes == 0:
            error_msg = _("size set to 'MAX' but could not allocate physical "
                          "disk space")
            raise exception.RedfishError(error=error_msg)

        logical_disk['size_bytes'] = max_volume_size_bytes
    elif max_volume_size_bytes < logical_disk['size_bytes']:
        error_msg = _('The physical disk space (%(max_vol_size)s bytes) is '
                      'not enough for the size of the logical disk '
                      '(%(logical_size)s bytes)' %
                      {'max_vol_size': max_volume_size_bytes,
                       'logical_size': logical_disk['size_bytes']})
        raise exception.RedfishError(error=error_msg)

    disk_usage = _volume_usage_per_disk_bytes(logical_disk, selected_disks,
                                              spans_count=spans_count)

    for disk in selected_disks:
        if free_space_bytes[disk] < disk_usage:
            error_msg = _('The free space of a disk (%(free_space)s bytes) '
                          'is not enough for the per-disk size of the logical '
                          'disk (%(disk_usage)s bytes)' %
                          {'free_space': free_space_bytes[disk],
                           'disk_usage': disk_usage})
            raise exception.RedfishError(error=error_msg)
        else:
            free_space_bytes[disk] -= disk_usage

    if 'controller' not in logical_disk:
        storage = disk_to_storage[selected_disks[0]]
        if storage:
            logical_disk['controller'] = storage.identity


def _raid_level_min_disks(raid_level, spans_count=1):
    try:
        raid_level_info = RAID_LEVELS[raid_level]
    except KeyError:
        _raise_raid_level_not_supported(raid_level)

    if raid_level_info['type'] == 'spanned':
        if spans_count <= 1:
            reason = _('Spanned RAID volumes cannot contain a single span')
            raise exception.RedfishError(error=reason)

        span_type = raid_level_info['span_type']
        raid_level_info = RAID_LEVELS[span_type]

    return raid_level_info['min_disks'] * spans_count


def _raid_level_max_disks(raid_level, spans_count=1):
    try:
        raid_level_info = RAID_LEVELS[raid_level]
    except KeyError:
        _raise_raid_level_not_supported(raid_level)

    if raid_level_info['type'] == 'spanned':
        if spans_count <= 1:
            reason = _('Spanned RAID volumes cannot contain a single span')
            raise exception.RedfishError(error=reason)

        span_type = raid_level_info['span_type']
        raid_level_info = RAID_LEVELS[span_type]

    return raid_level_info['max_disks'] * spans_count


def _usable_disks_count(raid_level, disks_count):
    """Calculates the number of disks usable for a RAID level

    ...given a physical disk count

    :param raid_level: RAID level of the virtual disk.
    :param disks_count: number of physical disks used for the virtual disk.
    :returns: number of disks.
    :raises: RedfishError if RAID level is not supported.
    """
    if raid_level in ['0', '1', '5', '6']:
        return disks_count
    elif raid_level in ['5+0', '6+0', '1+0']:
        # largest even number less than disk_count
        return (disks_count >> 1) << 1
    else:
        _raise_raid_level_not_supported(raid_level)


def _assign_disks_to_volume(logical_disks, physical_disks_by_type,
                            free_space_bytes, disk_to_storage):
    logical_disk = logical_disks.pop(0)
    raid_level = logical_disk['raid_level']

    # iterate over all possible configurations
    for (disk_type,
         protocol, size_bytes), disks in physical_disks_by_type.items():
        if ('disk_type' in logical_disk
                and logical_disk['disk_type'].lower() != disk_type.lower()):
            continue
        if ('interface_type' in logical_disk
                and logical_disk['interface_type'].lower()
                != PROTOCOL_MAP[protocol].lower()):
            continue

        # filter out disks without free disk space
        disks = [disk for disk in disks if free_space_bytes[disk] > 0]

        # sort disks by free size which is important if we have max disks limit
        # on a volume
        disks = sorted(
            disks,
            key=lambda disk: free_space_bytes[disk])

        # filter out disks already in use if sharing is disabled
        if ('share_physical_disks' not in logical_disk
                or not logical_disk['share_physical_disks']):
            disks = [disk for disk in disks
                     if disk.capacity_bytes == free_space_bytes[disk]]

        max_spans = _calculate_spans(raid_level, len(disks))
        min_spans = min([2, max_spans])
        min_disks = _raid_level_min_disks(raid_level,
                                          spans_count=min_spans)
        max_disks = _raid_level_max_disks(raid_level,
                                          spans_count=max_spans)
        candidate_max_disks = min([max_disks, len(disks)])

        for disks_count in range(min_disks, candidate_max_disks + 1):
            if ('number_of_physical_disks' in logical_disk
                    and logical_disk[
                        'number_of_physical_disks'] != disks_count):
                continue

            # skip invalid disks_count
            if disks_count != _usable_disks_count(logical_disk['raid_level'],
                                                  disks_count):
                continue

            selected_disks = disks[0:disks_count]

            candidate_volume = logical_disk.copy()
            candidate_free_space_bytes = free_space_bytes.copy()
            candidate_volume['physical_disks'] = [disk.identity for disk
                                                  in selected_disks]
            try:
                _calculate_volume_props(candidate_volume, selected_disks,
                                        candidate_free_space_bytes,
                                        disk_to_storage)
            except exception.RedfishError as exc:
                LOG.debug('Caught RedfishError in _calculate_volume_props(). '
                          'Reason: %s', exc)
                continue

            if len(logical_disks) > 0:
                try:
                    result, candidate_free_space_bytes = (
                        _assign_disks_to_volume(logical_disks,
                                                physical_disks_by_type,
                                                candidate_free_space_bytes,
                                                disk_to_storage))
                except exception.RedfishError as exc:
                    LOG.debug('Caught RedfishError in '
                              '_assign_disks_to_volume(). Reason: %s', exc)
                    continue
                if result:
                    logical_disks.append(candidate_volume)
                    return True, candidate_free_space_bytes
            else:
                logical_disks.append(candidate_volume)
                return True, candidate_free_space_bytes
    else:
        # put back the logical_disk to queue
        logical_disks.insert(0, logical_disk)
        return False, free_space_bytes


def _find_configuration(logical_disks, physical_disks, disk_to_storage):
    """Find RAID configuration.

    This method transforms the RAID configuration defined in Ironic to a format
    that is required by sushy. This includes matching the physical disks
    to RAID volumes when it's not pre-defined, or in general calculating
    missing properties.
    """

    # shared physical disks of RAID volumes size_gb='MAX' should be
    # de-prioritized during the matching process to reserve as much space as
    # possible. Reserved means it won't be used during matching.
    volumes_with_reserved_physical_disks = [
        volume for volume in logical_disks
        if ('physical_disks' in volume and volume['size_bytes'] == 'MAX'
            and volume.get('share_physical_disks', False))]
    reserved_physical_disks = [
        disk for disk in physical_disks
        for volume in volumes_with_reserved_physical_disks
        if disk.identity in volume['physical_disks']]

    # we require each logical disk contain only homogeneous physical disks, so
    # sort them by type
    physical_disks_by_type = {}
    reserved_physical_disks_by_type = {}
    free_space_bytes = {}
    for disk in physical_disks:
        # calculate free disk space
        # NOTE(billdodd): This won't be true if part of the drive is being used
        #     by an existing Volume, but has some space available for new
        #     Volumes. Redfish and/or SNIA may address this case in future.
        free_space_bytes[disk] = disk.capacity_bytes

        disk_type = (disk.media_type, disk.protocol, disk.capacity_bytes)
        if disk_type not in physical_disks_by_type:
            physical_disks_by_type[disk_type] = []
            reserved_physical_disks_by_type[disk_type] = []

        if disk in reserved_physical_disks:
            reserved_physical_disks_by_type[disk_type].append(disk)
        else:
            physical_disks_by_type[disk_type].append(disk)

    # exclude non-shared physical disks (predefined by the user) from
    # physical_disks_by_type because they are not going to be used during
    # matching
    for volume in logical_disks:
        if ('physical_disks' in volume
                and not volume.get('share_physical_disks', False)):
            for disk in physical_disks:
                if disk.identity in volume['physical_disks']:
                    disk_type = (disk.media_type, disk.protocol,
                                 disk.capacity_bytes)
                    if disk in physical_disks_by_type[disk_type]:
                        physical_disks_by_type[disk_type].remove(disk)

    processed_volumes = []

    # step 1 - process volumes with predefined disks and exact size
    for volume in [volume for volume in logical_disks
                   if ('physical_disks' in volume
                       and volume['size_bytes'] != 'MAX')]:
        _calculate_volume_props(volume, physical_disks, free_space_bytes,
                                disk_to_storage)
        processed_volumes.append(volume)

    # step 2 - process volumes without predefined disks
    volumes_without_disks = [disk for disk in logical_disks
                             if 'physical_disks' not in disk]

    if volumes_without_disks:
        result, free_space_bytes = (
            _assign_disks_to_volume(volumes_without_disks,
                                    physical_disks_by_type, free_space_bytes,
                                    disk_to_storage))
        if not result:
            # try again using the reserved physical disks in addition
            for disk_type, disks in physical_disks_by_type.items():
                physical_disks_by_type[disk_type] += (
                    reserved_physical_disks_by_type[disk_type])

            result, free_space_bytes = (
                _assign_disks_to_volume(volumes_without_disks,
                                        physical_disks_by_type,
                                        free_space_bytes,
                                        disk_to_storage))
            if not result:
                error_msg = _('failed to find matching physical disks for all '
                              'logical disks')
                LOG.error('Redfish driver failed to create RAID '
                          'configuration. Reason: %(error)s.',
                          {'error': error_msg})
                raise exception.RedfishError(error=error_msg)

    processed_volumes += volumes_without_disks

    # step 3 - process volumes with predefined disks and size_bytes == 'MAX'
    for volume in [volume for volume in logical_disks
                   if ('physical_disks' in volume
                       and volume['size_bytes'] == 'MAX')]:
        _calculate_volume_props(volume, physical_disks, free_space_bytes,
                                disk_to_storage)
        processed_volumes.append(volume)

    return processed_volumes


def _filter_logical_disks(logical_disks, include_root_volume,
                          include_nonroot_volumes):
    filtered_disks = []
    for disk in logical_disks:
        if include_root_volume and disk.get('is_root_volume'):
            filtered_disks.append(disk)

        if include_nonroot_volumes and not disk.get('is_root_volume'):
            filtered_disks.append(disk)

    return filtered_disks


def _get_storage_controller(node, system, physical_disks):
    collection = system.storage
    for storage in collection.get_members():
        # Using first controller as expecting only one
        controller = redfish_utils.get_first_controller(storage)
        if controller and controller.raid_types == []:
            continue
        for drive in storage.drives:
            if drive.identity in physical_disks:
                return storage


def _drive_path(storage, drive_id):
    for drive in storage.drives:
        if drive.identity == drive_id:
            return drive._path


def _construct_volume_payload(
        node, storage, raid_controller, physical_disks, raid_level, size_bytes,
        disk_name=None, span_length=None, span_depth=None):
    payload = {
        'RAIDType': RAID_LEVELS[raid_level]['raid_type'],
        'CapacityBytes': size_bytes
    }
    if physical_disks:
        payload['Links'] = {
            "Drives": [{"@odata.id": _drive_path(storage, d)} for d in
                       physical_disks]
        }
    if disk_name:
        payload['Name'] = disk_name
    LOG.debug('Payload for RAID logical disk creation on node %(node_uuid)s: '
              '%(payload)r', {'node_uuid': node.uuid, 'payload': payload})
    return payload


def create_virtual_disk(task, raid_controller, physical_disks, raid_level,
                        size_bytes, disk_name=None, span_length=None,
                        span_depth=None, error_handler=None):
    """Create a single virtual disk on a RAID controller.

    :param task: TaskManager object containing the node.
    :param raid_controller: id of the RAID controller.
    :param physical_disks: ids of the physical disks.
    :param raid_level: RAID level of the virtual disk.
    :param size_bytes: size of the virtual disk.
    :param disk_name: name of the virtual disk. (optional)
    :param span_depth: Number of spans in virtual disk. (optional)
    :param span_length: Number of disks per span. (optional)
    :param error_handler: function to call if volume create fails. (optional)
    :returns: Newly created Volume resource or TaskMonitor if async task.
    :raises: RedfishConnectionError when it fails to connect to Redfish.
    :raises: RedfishError if there is an error creating the virtual disk.
    """
    node = task.node
    system = redfish_utils.get_system(node)
    storage = _get_storage_controller(node, system, physical_disks)
    if not storage:
        reason = _('No storage controller found for node %(node_uuid)s' %
                   {'node_uuid': node.uuid})
        raise exception.RedfishError(error=reason)
    volume_collection = storage.volumes

    apply_time = None
    apply_time_support = volume_collection.operation_apply_time_support
    if apply_time_support and apply_time_support.mapped_supported_values:
        supported_values = apply_time_support.mapped_supported_values
        if sushy.APPLY_TIME_IMMEDIATE in supported_values:
            apply_time = sushy.APPLY_TIME_IMMEDIATE
        elif sushy.APPLY_TIME_ON_RESET in supported_values:
            apply_time = sushy.APPLY_TIME_ON_RESET
    payload = _construct_volume_payload(
        node, storage, raid_controller, physical_disks, raid_level, size_bytes,
        disk_name=disk_name, span_length=span_length, span_depth=span_depth)

    try:
        return volume_collection.create(payload, apply_time=apply_time)
    except sushy.exceptions.SushyError as exc:
        msg = ('Redfish driver failed to create virtual disk for node '
               '%(node_uuid)s. Reason: %(error)s.')
        if error_handler:
            try:
                return error_handler(task, exc, volume_collection, payload)
            except sushy.exceptions.SushyError as exc:
                LOG.error(msg, {'node_uuid': node.uuid, 'error': exc})
                raise exception.RedfishError(error=exc)
        LOG.error(msg, {'node_uuid': node.uuid, 'error': exc})
        raise exception.RedfishError(error=exc)


def update_raid_config(node):
    """Updates node's raid_config field with current logical disks.

    :param node: node for which to update the raid_config field
    """
    system = redfish_utils.get_system(node)
    logical_disks = []
    vol_no_raid_type = []
    for stor in system.storage.get_members():
        for vol in stor.volumes.get_members():
            if vol.raid_type:
                logical_disk = {
                    'id': vol.identity,
                    'name': vol.name,
                    'controller': stor.identity,
                    'size_gb': int(vol.capacity_bytes / units.Gi),
                    'raid_level': next(
                        key for key, value in RAID_LEVELS.items()
                        if value['raid_type'] == vol.raid_type.value)
                }
                logical_disks.append(logical_disk)
            else:
                vol_no_raid_type.append(vol.identity)

    if vol_no_raid_type:
        LOG.warning("Unable to update raid_config for volumes missing RAID "
                    "type: %(vol_no_raid_type)s",
                    {'vol_no_raid_type': ", ".join(vol_no_raid_type)})

    raid.update_raid_info(node, {'logical_disks': logical_disks})


class RedfishRAID(base.RAIDInterface):

    def get_properties(self):
        """Return the properties of the interface.

        :returns: dictionary of <property name>:<property description> entries.
        """
        return redfish_utils.COMMON_PROPERTIES.copy()

    def validate_raid_config(self, task, raid_config):
        """Validates the given RAID configuration.

        :param task: A TaskManager instance.
        :param raid_config: The RAID configuration to validate.
        :raises: InvalidParameterValue, if the RAID configuration is invalid.
        """

        super(RedfishRAID, self).validate_raid_config(task, raid_config)

        # Check if any interface_type is scsi that is not supported by Redfish
        scsi_disks = ([x for x in raid_config['logical_disks']
                      if x.get('interface_type') == raid.SCSI])

        if len(scsi_disks) > 0:
            raise exception.InvalidParameterValue(
                _('interface type `scsi` not supported by Redfish RAID'))

    @base.service_step(priority=0,
                       argsinfo=base.RAID_APPLY_CONFIGURATION_ARGSINFO)
    @base.deploy_step(priority=0,
                      argsinfo=base.RAID_APPLY_CONFIGURATION_ARGSINFO)
    def apply_configuration(self, task, raid_config, create_root_volume=True,
                            create_nonroot_volumes=False,
                            delete_existing=False):
        return super(RedfishRAID, self).apply_configuration(
            task, raid_config, create_root_volume=create_root_volume,
            create_nonroot_volumes=create_nonroot_volumes,
            delete_existing=delete_existing)

    @base.clean_step(priority=0, abortable=False, argsinfo={
        'create_root_volume': {
            'description': (
                'This specifies whether to create the root volume. '
                'Defaults to `True`.'
            ),
            'required': False
        },
        'create_nonroot_volumes': {
            'description': (
                'This specifies whether to create the non-root volumes. '
                'Defaults to `True`.'
            ),
            'required': False
        },
        'delete_existing': {
            'description': (
                'Setting this to `True` indicates to delete existing RAID '
                'configuration prior to creating the new configuration. '
                'Default value is `False`.'
            ),
            'required': False,
        }
    })
    def create_configuration(self, task, create_root_volume=True,
                             create_nonroot_volumes=True,
                             delete_existing=False):
        """Create RAID configuration on the node.

        This method creates the RAID configuration as read from
        node.target_raid_config.  This method
        by default will create all logical disks.

        :param task: TaskManager object containing the node.
        :param create_root_volume: Setting this to False indicates
            not to create root volume that is specified in the node's
            target_raid_config. Default value is True.
        :param create_nonroot_volumes: Setting this to False indicates
            not to create non-root volumes (all except the root volume) in
            the node's target_raid_config.  Default value is True.
        :param delete_existing: Setting this to True indicates to delete RAID
            configuration prior to creating the new configuration. Default is
            False.
        :returns: states.CLEANWAIT if RAID configuration is in progress
            asynchronously or None if it is complete.
        :raises: RedfishError if there is an error creating the configuration
        """
        node = task.node

        logical_disks = node.target_raid_config['logical_disks']
        convert_drive_units(logical_disks, node)
        physical_disks, disk_to_storage = get_physical_disks(node)
        # TODO(billdodd): filter out physical disks that are already in use?
        #                 filter out disks with HotSpareType != "None"?
        logical_disks = _find_configuration(logical_disks, physical_disks,
                                            disk_to_storage)

        logical_disks_to_create = _filter_logical_disks(
            logical_disks, create_root_volume, create_nonroot_volumes)

        logical_disks_to_create = self.pre_create_configuration(
            task, logical_disks_to_create)

        # Group logical disks by controller
        def gb_key(x):
            return x.get('controller')
        gb_list = itertools.groupby(
            sorted(logical_disks_to_create, key=gb_key), gb_key)
        ld_grouped = {k: list(g) for k, g in gb_list}
        raid_configs, reboot_required = self._submit_create_configuration(
            task, ld_grouped)

        return_state = None
        deploy_utils.set_async_step_flags(
            node,
            reboot=reboot_required,
            skip_current_step=True,
            polling=True)
        if reboot_required:
            return_state = deploy_utils.reboot_to_finish_step(task)
        else:
            update_raid_config(node)

        return self.post_create_configuration(
            task, raid_configs, return_state=return_state)

    @base.clean_step(priority=0)
    @base.deploy_step(priority=0)
    @base.service_step(priority=0)
    def delete_configuration(self, task):
        """Delete RAID configuration on the node.

        :param task: TaskManager object containing the node.
        :returns: states.CLEANWAIT (cleaning) or states.DEPLOYWAIT (deployment)
            if deletion is in progress asynchronously or None if it is
            complete.
        """
        node = task.node
        raid_configs, reboot_required = self._submit_delete_configuration(
            task)

        return_state = None
        deploy_utils.set_async_step_flags(
            node,
            reboot=reboot_required,
            skip_current_step=True,
            polling=True)
        if reboot_required:
            return_state = deploy_utils.reboot_to_finish_step(task)
        else:
            update_raid_config(node)

        return self.post_delete_configuration(
            task, raid_configs, return_state=return_state)

    def volume_create_error_handler(self, task, exc, volume_collection,
                                    payload):
        """Handle error from failed VolumeCollection.create()

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action if the call to
        VolumeCollection.create() fails.

        :param task: a TaskManager instance containing the node to act on.
        :param exc: the exception raised by VolumeCollection.create().
        :param volume_collection: the sushy VolumeCollection instance.
        :param payload: the payload passed to the failed create().
        :returns: Newly created Volume resource or TaskMonitor if async task.
        :raises: RedfishError if there is an error creating the virtual disk.
        """
        raise exc

    def pre_create_configuration(self, task, logical_disks_to_create):
        """Perform required actions before creating config.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform custom actions prior to creating
        the RAID configuration on the Redfish service.

        :param task: a TaskManager instance containing the node to act on.
        :param logical_disks_to_create: list of logical disks to create.
        :returns: updated list of logical disks to create.
        """
        return logical_disks_to_create

    def post_create_configuration(self, task, raid_configs, return_state=None):
        """Perform post create_configuration action to commit the config.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action to commit the
        RAID create configuration to the Redfish service.

        :param task: a TaskManager instance containing the node to act on.
        :param raid_configs: a list of dictionaries containing the RAID
                             configuration operation details.
        :param return_state: state to return based on operation being invoked
        """
        return return_state

    def pre_delete_configuration(self, task, vols_to_delete):
        """Perform required actions before deleting config.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform custom actions prior to deleting
        the RAID configuration on the Redfish service.

        :param task: a TaskManager instance containing the node to act on.
        :param vols_to_delete: list of volumes to delete.
        """
        pass

    def post_delete_configuration(self, task, raid_configs, return_state=None):
        """Perform post delete_configuration action to commit the config.

        Extension point to allow vendor implementations to extend this class
        and override this method to perform a custom action to commit the
        RAID delete configuration to the Redfish service.

        :param task: a TaskManager instance containing the node to act on.
        :param raid_configs: a list of dictionaries containing the RAID
                             configuration operation details.
        :param return_state: state to return based on operation being invoked
        """
        return return_state

    def _clear_raid_configs(self, node):
        """Clears RAID configurations from driver_internal_info

        Note that the caller must have an exclusive lock on the node.

        :param node: the node to clear the RAID configs from
        """
        node.del_driver_internal_info('raid_configs')
        node.save()

    @METRICS.timer('RedfishRAID._query_raid_config_failed')
    @periodics.node_periodic(
        purpose='checking async RAID config failed',
        spacing=CONF.redfish.raid_config_fail_interval,
        filters={'reserved': False, 'provision_state_in': {
            states.CLEANFAIL, states.DEPLOYFAIL}, 'maintenance': True},
        predicate_extra_fields=['driver_internal_info'],
        predicate=lambda n: n.driver_internal_info.get('raid_configs'),
    )
    def _query_raid_config_failed(self, task, manager, context):
        """Periodic job to check for failed RAID configuration."""
        # A RAID config failed. Discard any remaining RAID
        # configs so when the user takes the node out of
        # maintenance mode, pending RAID configs do not
        # automatically continue.
        LOG.warning('RAID configuration failed for node %(node)s. '
                    'Discarding remaining RAID configurations.',
                    {'node': task.node.uuid})

        task.upgrade_lock()
        self._clear_raid_configs(task.node)
        update_raid_config(task.node)

    @METRICS.timer('RedfishRAID._query_raid_config_status')
    @periodics.node_periodic(
        purpose='checking async RAID config tasks',
        spacing=CONF.redfish.raid_config_status_interval,
        filters={'reserved': False, 'provision_state_in': {
            states.CLEANWAIT, states.DEPLOYWAIT}},
        predicate_extra_fields=['driver_internal_info'],
        predicate=lambda n: n.driver_internal_info.get('raid_configs'),
    )
    def _query_raid_config_status(self, task, manager, context):
        """Periodic job to check RAID config tasks."""
        self._check_node_raid_config(task)

    def _raid_config_in_progress(self, task, task_monitor_uri, operation):
        """Check if this RAID configuration operation is still in progress.

        :param task: TaskManager object containing the node.
        :param task_monitor_uri: Redfish task monitor URI
        :param operation: 'create' or 'delete' operation for given task.
            Used in log messages.
        :returns: True, if still in progress, otherwise False.
        """
        try:
            task_monitor = redfish_utils.get_task_monitor(task.node,
                                                          task_monitor_uri)
        except exception.RedfishError:
            LOG.info('Unable to get status of RAID %(operation)s task '
                     '%(task_mon_uri)s to node %(node_uuid)s; assuming task '
                     'completed successfully',
                     {'operation': operation,
                      'task_mon_uri': task_monitor_uri,
                      'node_uuid': task.node.uuid})
            return False
        if task_monitor.is_processing:
            LOG.debug('RAID %(operation)s task %(task_mon_uri)s to node '
                      '%(node_uuid)s still in progress',
                      {'operation': operation,
                       'task_mon_uri': task_monitor.task_monitor_uri,
                       'node_uuid': task.node.uuid})
            return True
        else:
            sushy_task = task_monitor.get_task()
            messages = []
            if sushy_task.messages and not sushy_task.messages[0].message:
                sushy_task.parse_messages()

            messages = [m.message for m in sushy_task.messages]

            if (sushy_task.task_state == sushy.TASK_STATE_COMPLETED
                    and sushy_task.task_status in
                    [sushy.HEALTH_OK, sushy.HEALTH_WARNING]):
                LOG.info('RAID %(operation)s task %(task_mon_uri)s to node '
                         '%(node_uuid)s completed.',
                         {'operation': operation,
                          'task_mon_uri': task_monitor.task_monitor_uri,
                          'node_uuid': task.node.uuid})
            else:
                LOG.error('RAID %(operation)s task %(task_mon_uri)s to node '
                          '%(node_uuid)s failed; messages: %(messages)s',
                          {'operation': operation,
                           'task_mon_uri': task_monitor.task_monitor_uri,
                           'node_uuid': task.node.uuid,
                           'messages': ", ".join(messages)})
        return False

    @METRICS.timer('RedfishRAID._check_node_raid_config')
    def _check_node_raid_config(self, task):
        """Check the progress of running RAID config on a node.

        :param task: TaskManager object containing the node.
        """
        node = task.node
        raid_configs = node.driver_internal_info['raid_configs']

        task.upgrade_lock()
        raid_configs['task_monitor_uri'] =\
            [i for i in raid_configs.get('task_monitor_uri')
             if self._raid_config_in_progress(
                task, i, raid_configs.get('operation'))]
        node.set_driver_internal_info('raid_configs', raid_configs)

        if not raid_configs['task_monitor_uri']:
            if raid_configs.get('pending'):
                if raid_configs.get('operation') == 'create':
                    reboot_required = self._submit_create_configuration(
                        task, raid_configs.get('pending'))[1]
                else:
                    reboot_required = self._submit_delete_configuration(
                        task)[1]
                if reboot_required:
                    deploy_utils.reboot_to_finish_step(task)
            else:
                self._clear_raid_configs(node)
                LOG.info('RAID configuration completed for node %(node)s',
                         {'node': node.uuid})
                update_raid_config(task.node)
                if task.node.clean_step:
                    manager_utils.notify_conductor_resume_clean(task)
                else:
                    manager_utils.notify_conductor_resume_deploy(task)

    def _submit_create_configuration(self, task, ld_grouped):
        """Processes and submits requests for creating RAID configuration.

        :param task: TaskManager object containing the node.
        :param ld_grouped: Dictionary of logical disks, grouped by controller.

        :returns: tuple of 1) dictionary containing operation name (create),
            pending items, and task monitor URIs, and 2) flag indicating if
            reboot is required.
        """
        node = task.node
        reboot_required = False
        raid_configs = {'operation': 'create', 'pending': {}}
        for controller, logical_disks in ld_grouped.items():
            iter_logical_disks = iter(logical_disks)
            for logical_disk in iter_logical_disks:
                response = create_virtual_disk(
                    task,
                    raid_controller=logical_disk.get('controller'),
                    physical_disks=logical_disk['physical_disks'],
                    raid_level=logical_disk['raid_level'],
                    size_bytes=logical_disk['size_bytes'],
                    disk_name=logical_disk.get('volume_name'),
                    span_length=logical_disk.get('span_length'),
                    span_depth=logical_disk.get('span_depth'),
                    error_handler=self.volume_create_error_handler)
                if (response is not None
                        and hasattr(response, 'task_monitor_uri')):
                    raid_configs.setdefault('task_monitor_uri', []).append(
                        response.task_monitor_uri)
                    reboot_required = True
                    # Don't process any on this controller until these
                    # created to avoid failures where only 1 request
                    # per controller can be submitted for non-immediate
                    break
            # Append remaining disks for this controller, if any left
            for logical_disk in iter_logical_disks:
                raid_configs['pending'].setdefault(controller, []).append(
                    logical_disk)

        # Store only when async operation
        if reboot_required:
            node.set_driver_internal_info('raid_configs', raid_configs)

        return raid_configs, reboot_required

    def _submit_delete_configuration(self, task):
        """Processes and submits requests for deleting virtual disks.

        :param task: TaskManager object containing the node.

        :returns: tuple of 1) dictionary containing operation name (delete),
            flag to indicate if any disks remaining, and task monitor URIs,
            and 2) flag indicating if reboot is required
        :raises RedfishError: if fails to get list of virtual disks
        """
        node = task.node
        system = redfish_utils.get_system(node)
        vols_to_delete = {}
        any_left = False
        try:
            for storage in system.storage.get_members():
                controller = redfish_utils.get_first_controller(storage)
                controller_id = None
                if controller:
                    controller_id = storage.identity
                iter_volumes = iter(storage.volumes.get_members())
                for volume in iter_volumes:
                    if volume.raid_type:
                        if controller_id not in vols_to_delete:
                            vols_to_delete[controller_id] = []
                        apply_time = self._get_apply_time(
                            storage.volumes.operation_apply_time_support)
                        vols_to_delete[controller_id].append((
                            apply_time, volume))
                        if apply_time == sushy.APPLY_TIME_ON_RESET:
                            # Don't process any on this controller until these
                            # deleted to avoid failures where only 1 request
                            # per controller can be submitted for non-immediate
                            break
                any_left = any(iter_volumes)
        except sushy.exceptions.SushyError as exc:
            error_msg = _('Cannot get the list of volumes to delete for node '
                          '%(node_uuid)s. Reason: %(error)s.' %
                          {'node_uuid': node.uuid, 'error': exc})
            LOG.error(error_msg)
            raise exception.RedfishError(error=exc)

        self.pre_delete_configuration(task, vols_to_delete)

        reboot_required = False
        raid_configs = {'operation': 'delete', 'pending': any_left}
        for controller, vols_to_delete in vols_to_delete.items():
            for apply_time, volume in vols_to_delete:
                response = volume.delete(apply_time=apply_time)
                # only save the async tasks (task_monitors) in raid_config
                if (response is not None
                        and hasattr(response, 'task_monitor_uri')):
                    raid_configs.setdefault('task_monitor_uri', []).append(
                        response.task_monitor_uri)
                    reboot_required = True

        # Store only when async operation
        if reboot_required:
            node.set_driver_internal_info('raid_configs', raid_configs)

        return raid_configs, reboot_required

    def _get_apply_time(self, apply_time_support):
        """Gets apply time for RAID operations

        :param apply_time_support: Supported apply times
        :returns: None, if supported apply times not specified. Otherwise
            Immediate when available, or OnReset that will require rebooting.
        """
        apply_time = None
        if apply_time_support and apply_time_support.mapped_supported_values:
            supported_values = apply_time_support.mapped_supported_values
            if sushy.APPLY_TIME_IMMEDIATE in supported_values:
                apply_time = sushy.APPLY_TIME_IMMEDIATE
            elif sushy.APPLY_TIME_ON_RESET in supported_values:
                apply_time = sushy.APPLY_TIME_ON_RESET
        return apply_time



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

from oslo_concurrency import processutils
from oslo_utils import units
import tenacity

from ironic.common import utils
from ironic.conf import CONF


LOG = logging.getLogger(__name__)

# Limit the memory address space to 1 GiB when running qemu-img
QEMU_IMG_LIMITS = None


def _qemu_img_limits():
    # NOTE(TheJulia): If you make *any* chance to this code, you may need
    # to make an identitical or similar change to ironic-python-agent.
    global QEMU_IMG_LIMITS
    if QEMU_IMG_LIMITS is None:
        QEMU_IMG_LIMITS = processutils.ProcessLimits(
            address_space=CONF.disk_utils.image_convert_memory_limit
            * units.Mi)
    return QEMU_IMG_LIMITS


def _retry_on_res_temp_unavailable(exc):
    if (isinstance(exc, processutils.ProcessExecutionError)
            and ('Resource temporarily unavailable' in exc.stderr
                 or 'Cannot allocate memory' in exc.stderr)):
        return True
    return False


@tenacity.retry(
    retry=tenacity.retry_if_exception(_retry_on_res_temp_unavailable),
    stop=tenacity.stop_after_attempt(CONF.disk_utils.image_convert_attempts),
    reraise=True)
def convert_image(source, dest, out_format, run_as_root=False, cache=None,
                  out_of_order=False, sparse_size=None,
                  source_format='qcow2'):
    # NOTE(TheJulia): If you make *any* chance to this code, you may need
    # to make an identitical or similar change to ironic-python-agent.
    """Convert image to other format."""
    cmd = ['qemu-img', 'convert', '-f', source_format, '-O', out_format]
    if cache is not None:
        cmd += ['-t', cache]
    if sparse_size is not None:
        cmd += ['-S', sparse_size]
    if out_of_order:
        cmd.append('-W')
    cmd += [source, dest]
    # NOTE(TheJulia): Statically set the MALLOC_ARENA_MAX to prevent leaking
    # and the creation of new malloc arenas which will consume the system
    # memory. If limited to 1, qemu-img consumes ~250 MB of RAM, but when
    # another thread tries to access a locked section of memory in use with
    # another thread, then by default a new malloc arena is created,
    # which essentially balloons the memory requirement of the machine.
    # Default for qemu-img is 8 * nCPU * ~250MB (based on defaults +
    # thread/code/process/library overhead. In other words, 64 GB. Limiting
    # this to 3 keeps the memory utilization in happy cases below the overall
    # threshold which is in place in case a malicious image is attempted to
    # be passed through qemu-img.
    env_vars = {'MALLOC_ARENA_MAX': '3'}
    try:
        utils.execute(*cmd, run_as_root=run_as_root,
                      prlimit=_qemu_img_limits(),
                      use_standard_locale=True,
                      env_variables=env_vars)
    except processutils.ProcessExecutionError as e:
        if ('Resource temporarily unavailable' in e.stderr
            or 'Cannot allocate memory' in e.stderr):
            LOG.debug('Failed to convert image, retrying. Error: %s', e)
            # Sync disk caches before the next attempt
            utils.execute('sync')
        raise



# -*- encoding: utf-8 -*-
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The Ironic Service API."""

import sys

from oslo_config import cfg
from oslo_log import log

from ironic.common import service as ironic_service
from ironic.common import wsgi_service

CONF = cfg.CONF

LOG = log.getLogger(__name__)


def _error_about_sqlite_usage():
    if 'sqlite' in CONF.database.connection.lower():
        # TODO(TheJulia): Make this a hard error in C*
        LOG.error('We have detected the API is being launched with a SQLite '
                  'database backend. This is unsupported, and will be a hard '
                  'error in the future. This is because multi-process use of '
                  'a single SQLite database is problematic in terms of '
                  'locking. A single process ironic model exists for use with '
                  'SQLite.')


def main():
    # Parse config file and command line options, then start logging
    ironic_service.prepare_service('ironic_api', sys.argv)
    ironic_service.ensure_rpc_transport()

    # Build and start the WSGI app
    launcher = ironic_service.process_launcher()
    server = wsgi_service.WSGIService('ironic_api', CONF.api.enable_ssl_api)
    launcher.launch_service(server, workers=server.workers)
    sys.exit(launcher.wait())


if __name__ == '__main__':
    _error_about_sqlite_usage()
    sys.exit(main())



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii

from construct import core
import netaddr
from oslo_log import log as logging

from ironic.common import exception
from ironic.drivers.modules.inspector.hooks import base
from ironic.drivers.modules.inspector import lldp_tlvs as tlv
import ironic.objects.port as ironic_port


LOG = logging.getLogger(__name__)
PORT_ID_ITEM_NAME = "port_id"
SWITCH_ID_ITEM_NAME = "switch_id"


class LocalLinkConnectionHook(base.InspectionHook):
    """Hook to process mandatory LLDP packet fields"""

    dependencies = ['validate-interfaces']

    def _get_local_link_patch(self, lldp_data, port, node_uuid):
        local_link_connection = {}

        for tlv_type, tlv_value in lldp_data:
            try:
                data = bytearray(binascii.unhexlify(tlv_value))
            except binascii.Error:
                LOG.warning('TLV value for TLV type %d is not in correct '
                            'format. Ensure that the TLV value is in '
                            'hexadecimal format when sent to ironic. Node: %s',
                            tlv_type, node_uuid)
                return

            item = value = None
            if tlv_type == tlv.LLDP_TLV_PORT_ID:
                try:
                    port_id = tlv.PortId.parse(data)
                except (core.MappingError, netaddr.AddrFormatError) as e:
                    LOG.warning('TLV parse error for Port ID for node %s: %s',
                                node_uuid, e)
                    return

                item = PORT_ID_ITEM_NAME
                value = port_id.value.value if port_id.value else None
            elif tlv_type == tlv.LLDP_TLV_CHASSIS_ID:
                try:
                    chassis_id = tlv.ChassisId.parse(data)
                except (core.MappingError, netaddr.AddrFormatError) as e:
                    LOG.warning('TLV parse error for Chassis ID for node %s: '
                                '%s', node_uuid, e)
                    return

                # Only accept mac address for chassis ID
                if 'mac_address' in chassis_id.subtype:
                    item = SWITCH_ID_ITEM_NAME
                    value = chassis_id.value.value

            if item is None or value is None:
                continue
            if item in port.local_link_connection:
                continue
            local_link_connection[item] = value

        try:
            LOG.debug('Updating port %s for node %s', port.address, node_uuid)
            for item in local_link_connection:
                port.set_local_link_connection(item,
                                               local_link_connection[item])
            port.save()
        except exception.IronicException as e:
            LOG.warning('Failed to update port %(uuid)s for node %(node)s. '
                        'Error: %(error)s', {'uuid': port.id,
                                             'node': node_uuid,
                                             'error': e})

    def __call__(self, task, inventory, plugin_data):
        """Process LLDP data and patch Ironic port local link connection.

        Process the non-vendor-specific LLDP packet fields for each NIC found
        for a baremetal node, port ID and chassis ID. These fields, if found
        and if valid, will be saved into the local link connection information
        (port id and switch id) fields on the Ironic port that represents that
        NIC.
        """
        lldp_raw = plugin_data.get('lldp_raw') or {}

        for iface in inventory['interfaces']:
            # The all_interfaces field in plugin_data is provided by the
            # validate-interfaces hook, so it is a dependency for this hook (?)
            if iface['name'] not in plugin_data.get('all_interfaces'):
                continue

            mac_address = iface['mac_address']
            port = ironic_port.Port.get_by_address(task.context, mac_address)
            if not port:
                LOG.debug('Skipping LLDP processing for interface %s of node '
                          '%s: matching port not found in Ironic.',
                          mac_address, task.node.uuid)
                continue

            lldp_data = lldp_raw.get(iface['name']) or iface.get('lldp')
            if lldp_data is None:
                LOG.warning('No LLDP data found for interface %s of node %s',
                            mac_address, task.node.uuid)
                continue

            # Parse raw lldp data
            self._get_local_link_patch(lldp_data, port, task.node.uuid)



# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Ironic test utilities."""


from oslo_utils import timeutils
from oslo_utils import uuidutils

from ironic.common import states
from ironic.db import api as db_api
from ironic.drivers import base as drivers_base
from ironic.objects import allocation
from ironic.objects import bios
from ironic.objects import chassis
from ironic.objects import conductor
from ironic.objects import deploy_template
from ironic.objects import firmware
from ironic.objects import inspection_rule as rule
from ironic.objects import node
from ironic.objects import node_history
from ironic.objects import node_inventory
from ironic.objects import port
from ironic.objects import portgroup
from ironic.objects import runbook
from ironic.objects import trait
from ironic.objects import volume_connector
from ironic.objects import volume_target


def get_test_ipmi_info():
    return {
        "ipmi_address": "1.2.3.4",
        "ipmi_username": "admin",
        "ipmi_password": "fake"
    }


def get_test_ipmi_bridging_parameters():
    return {
        "ipmi_bridging": "dual",
        "ipmi_local_address": "0x20",
        "ipmi_transit_channel": "0",
        "ipmi_transit_address": "0x82",
        "ipmi_target_channel": "7",
        "ipmi_target_address": "0x72"
    }


def get_test_pxe_driver_info():
    return {
        "deploy_kernel": "glance://deploy_kernel_uuid",
        "deploy_ramdisk": "glance://deploy_ramdisk_uuid",
        "rescue_kernel": "glance://rescue_kernel_uuid",
        "rescue_ramdisk": "glance://rescue_ramdisk_uuid"
    }


def get_test_pxe_driver_internal_info():
    return {
        "is_whole_disk_image": False,
    }


def get_test_pxe_instance_info():
    return {
        "image_source": "glance://image_uuid",
        "root_gb": 100,
        "rescue_password": "password"
    }


def get_test_ilo_info():
    return {
        "ilo_address": "1.2.3.4",
        "ilo_username": "admin",
        "ilo_password": "fake",
    }


def get_test_drac_info():
    return {
        "redfish_address": "1.2.3.4",
        "redfish_system_id": "/redfish/v1/Systems/System.Embedded.1",
        "redfish_username": "admin",
        "redfish_password": "fake"
    }


def get_test_irmc_info():
    return {
        "irmc_address": "1.2.3.4",
        "irmc_username": "admin0",
        "irmc_password": "fake0",
        "irmc_port": "80",
        "irmc_auth_method": "digest",
    }


def get_test_agent_instance_info():
    return {
        'image_source': 'fake-image',
        'image_url': 'http://image',
        'image_checksum': 'checksum',
        'image_disk_format': 'qcow2',
        'image_container_format': 'bare',
    }


def get_test_agent_driver_info():
    return {
        'deploy_kernel': 'glance://deploy_kernel_uuid',
        'deploy_ramdisk': 'glance://deploy_ramdisk_uuid',
        'ipmi_password': 'foo',
    }


def get_test_agent_driver_internal_info():
    return {
        'agent_url': 'http://127.0.0.1/foo',
        'is_whole_disk_image': True,
    }


def get_test_snmp_info(**kw):
    result = {
        "snmp_driver": kw.get("snmp_driver", "teltronix"),
        "snmp_address": kw.get("snmp_address", "1.2.3.4"),
        "snmp_port": kw.get("snmp_port", "161"),
        "snmp_outlet": kw.get("snmp_outlet", "1"),
        "snmp_version": kw.get("snmp_version", "1")
    }
    if result["snmp_version"] in ("1", "2c"):
        result["snmp_community"] = kw.get("snmp_community", "public")
        if "snmp_community_read" in kw:
            result["snmp_community_read"] = kw["snmp_community_read"]
        if "snmp_community_write" in kw:
            result["snmp_community_write"] = kw["snmp_community_write"]
    elif result["snmp_version"] == "3":
        result["snmp_user"] = kw.get(
            "snmp_user", kw.get("snmp_security", "snmpuser")
        )
        for option in ('snmp_auth_protocol', 'snmp_auth_key',
                       'snmp_priv_protocol', 'snmp_priv_key',
                       'snmp_context_engine_id', 'snmp_context_name'):
            if option in kw:
                result[option] = kw[option]
    return result


def get_test_node(**kw):
    properties = {
        "cpu_arch": "x86_64",
        "local_gb": "10",
        "memory_mb": "4096",
    }
    # NOTE(tenbrae): API unit tests confirm that sensitive fields in
    #                instance_info and driver_info will get scrubbed
    #                from the API response but other fields
    #                (eg, 'foo') do not.
    fake_instance_info = {
        "configdrive": "TG9yZW0gaXBzdW0gZG9sb3Igc2l0IGFtZXQ=",
        "image_url": "http://example.com/test_image_url",
        "foo": "bar",
    }
    fake_driver_info = {
        "foo": "bar",
        "fake_password": "fakepass",
    }
    fake_internal_info = {
        "private_state": "secret value"
    }
    result = {
        'version': kw.get('version', node.Node.VERSION),
        'id': kw.get('id', 123),
        'name': kw.get('name', None),
        'uuid': kw.get('uuid', '1be26c0b-03f2-4d2e-ae87-c02d7f33c123'),
        'chassis_id': kw.get('chassis_id', None),
        'conductor_affinity': kw.get('conductor_affinity', None),
        'conductor_group': kw.get('conductor_group', ''),
        'power_state': kw.get('power_state', states.NOSTATE),
        'target_power_state': kw.get('target_power_state', states.NOSTATE),
        'provision_state': kw.get('provision_state', states.AVAILABLE),
        'target_provision_state': kw.get('target_provision_state',
                                         states.NOSTATE),
        'provision_updated_at': kw.get('provision_updated_at'),
        'last_error': kw.get('last_error'),
        'instance_uuid': kw.get('instance_uuid'),
        'instance_info': kw.get('instance_info', fake_instance_info),
        'driver': kw.get('driver', 'fake-hardware'),
        'driver_info': kw.get('driver_info', fake_driver_info),
        'driver_internal_info': kw.get('driver_internal_info',
                                       fake_internal_info),
        'clean_step': kw.get('clean_step'),
        'deploy_step': kw.get('deploy_step'),
        'properties': kw.get('properties', properties),
        'reservation': kw.get('reservation'),
        'maintenance': kw.get('maintenance', False),
        'maintenance_reason': kw.get('maintenance_reason'),
        'fault': kw.get('fault'),
        'console_enabled': kw.get('console_enabled', False),
        'extra': kw.get('extra', {}),
        'updated_at': kw.get('updated_at'),
        'created_at': kw.get('created_at'),
        'inspection_finished_at': kw.get('inspection_finished_at'),
        'inspection_started_at': kw.get('inspection_started_at'),
        'raid_config': kw.get('raid_config'),
        'target_raid_config': kw.get('target_raid_config'),
        'tags': kw.get('tags', []),
        'resource_class': kw.get('resource_class'),
        'traits': kw.get('traits', []),
        'automated_clean': kw.get('automated_clean', None),
        'protected': kw.get('protected', False),
        'protected_reason': kw.get('protected_reason', None),
        'conductor': kw.get('conductor'),
        'owner': kw.get('owner', None),
        'allocation_id': kw.get('allocation_id'),
        'description': kw.get('description'),
        'retired': kw.get('retired', False),
        'retired_reason': kw.get('retired_reason', None),
        'lessee': kw.get('lessee', None),
        'network_data': kw.get('network_data'),
        'boot_mode': kw.get('boot_mode', None),
        'secure_boot': kw.get('secure_boot', None),
        'shard': kw.get('shard', None),
        'parent_node': kw.get('parent_node', None),
        'service_step': kw.get('service_step'),
        'disable_power_off': kw.get('disable_power_off', False),
    }

    for iface in drivers_base.ALL_INTERFACES:
        name = '%s_interface' % iface
        result[name] = kw.get(name)

    return result


def create_test_node(**kw):
    """Create test node entry in DB and return Node DB object.

    Function to be used to create test Node objects in the database.

    :param kw: kwargs with overriding values for node's attributes.
    :returns: Test Node DB object.

    """
    node = get_test_node(**kw)
    # Let DB generate an ID if one isn't specified explicitly.
    # Creating a node with tags or traits will raise an exception. If tags or
    # traits are not specified explicitly just delete them.
    for field in {'id', 'uuid', 'tags', 'traits'}:
        if field not in kw:
            del node[field]
    dbapi = db_api.get_instance()
    return dbapi.create_node(node)


def get_test_port(**kw):
    return {
        'id': kw.get('id', 987),
        'version': kw.get('version', port.Port.VERSION),
        'uuid': kw.get('uuid', '1be26c0b-03f2-4d2e-ae87-c02d7f33c781'),
        'node_id': kw.get('node_id', 123),
        'node_uuid': kw.get('node_uuid',
                            '59d102f7-5840-4299-8ec8-80c0ebae9de1'),
        'address': kw.get('address', '52:54:00:cf:2d:31'),
        'extra': kw.get('extra', {}),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
        'local_link_connection': kw.get('local_link_connection',
                                        {'switch_id': '0a:1b:2c:3d:4e:5f',
                                         'port_id': 'Ethernet3/1',
                                         'switch_info': 'switch1'}),
        'portgroup_id': kw.get('portgroup_id'),
        'pxe_enabled': kw.get('pxe_enabled', True),
        'internal_info': kw.get('internal_info', {"bar": "buzz"}),
        'physical_network': kw.get('physical_network'),
        'is_smartnic': kw.get('is_smartnic', False),
        'name': kw.get('name'),
    }


def create_test_port(**kw):
    """Create test port entry in DB and return Port DB object.

    Function to be used to create test Port objects in the database.

    :param kw: kwargs with overriding values for port's attributes.
    :returns: Test Port DB object.

    """
    port = get_test_port(**kw)
    # Let DB generate ID if it isn't specified explicitly
    for field in ('id', 'uuid'):
        if field not in kw:
            del port[field]
    dbapi = db_api.get_instance()
    return dbapi.create_port(port)


def get_test_volume_connector(**kw):
    return {
        'id': kw.get('id', 789),
        'version': kw.get('version', volume_connector.VolumeConnector.VERSION),
        'uuid': kw.get('uuid', '1be26c0b-03f2-4d2e-ae87-c02d7f33c781'),
        'node_id': kw.get('node_id', 123),
        'type': kw.get('type', 'iqn'),
        'connector_id': kw.get('connector_id',
                               'iqn.2012-06.com.example:initiator'),
        'extra': kw.get('extra', {}),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_volume_connector(**kw):
    """Create test connector entry in DB and return VolumeConnector DB object.

    Function to be used to create test VolumeConnector objects in the database.

    :param kw: kwargs with overriding values for connector's attributes.
    :returns: Test VolumeConnector DB object.

    """
    connector = get_test_volume_connector(**kw)
    # Let DB generate ID if it isn't specified explicitly
    for field in ('id', 'uuid'):
        if field not in kw:
            del connector[field]
    dbapi = db_api.get_instance()
    return dbapi.create_volume_connector(connector)


def get_test_volume_target(**kw):
    fake_properties = {"target_iqn": "iqn.foo"}
    return {
        'id': kw.get('id', 789),
        'version': kw.get('version', volume_target.VolumeTarget.VERSION),
        'uuid': kw.get('uuid', '1be26c0b-03f2-4d2e-ae87-c02d7f33c781'),
        'node_id': kw.get('node_id', 123),
        'volume_type': kw.get('volume_type', 'iscsi'),
        'properties': kw.get('properties', fake_properties),
        'boot_index': kw.get('boot_index', 0),
        'volume_id': kw.get('volume_id', '12345678'),
        'extra': kw.get('extra', {}),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_volume_target(**kw):
    """Create test target entry in DB and return VolumeTarget DB object.

    Function to be used to create test VolumeTarget objects in the database.

    :param kw: kwargs with overriding values for target's attributes.
    :returns: Test VolumeTarget DB object.

    """
    target = get_test_volume_target(**kw)
    # Let DB generate ID if it isn't specified explicitly
    for field in ('id', 'uuid'):
        if field not in kw:
            del target[field]
    dbapi = db_api.get_instance()
    return dbapi.create_volume_target(target)


def get_test_chassis(**kw):
    return {
        'id': kw.get('id', 42),
        'version': kw.get('version', chassis.Chassis.VERSION),
        'uuid': kw.get('uuid', 'e74c40e0-d825-11e2-a28f-0800200c9a66'),
        'extra': kw.get('extra', {}),
        'description': kw.get('description', 'data-center-1-chassis'),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_chassis(**kw):
    """Create test chassis entry in DB and return Chassis DB object.

    Function to be used to create test Chassis objects in the database.

    :param kw: kwargs with overriding values for chassis's attributes.
    :returns: Test Chassis DB object.

    """
    chassis = get_test_chassis(**kw)
    # Let DB generate ID if it isn't specified explicitly
    for field in ('id', 'uuid'):
        if field not in kw:
            del chassis[field]
    dbapi = db_api.get_instance()
    return dbapi.create_chassis(chassis)


def get_test_conductor(**kw):
    return {
        'id': kw.get('id', 6),
        'version': kw.get('version', conductor.Conductor.VERSION),
        'hostname': kw.get('hostname', 'test-conductor-node'),
        'drivers': kw.get('drivers', ['fake-driver', 'null-driver']),
        'conductor_group': kw.get('conductor_group', ''),
        'created_at': kw.get('created_at', timeutils.utcnow()),
        'updated_at': kw.get('updated_at', timeutils.utcnow()),
        'online': kw.get('online', True)
    }


def create_test_conductor(**kw):
    """Create test conductor entry in DB and return Conductor DB object.

    Function to be used to create test Conductor objects in the database.

    :param kw: kwargs with overriding values for conductor's attributes.
    :returns: Test Conductor DB object.

    """
    conductor = get_test_conductor(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del conductor['id']
    dbapi = db_api.get_instance()
    return dbapi.register_conductor(conductor)


def get_test_redfish_info():
    return {
        "redfish_address": "https://example.com",
        "redfish_system_id": "/redfish/v1/Systems/FAKESYSTEM",
        "redfish_username": "username",
        "redfish_password": "password"
    }


def get_test_portgroup(**kw):
    return {
        'id': kw.get('id', 654),
        'version': kw.get('version', portgroup.Portgroup.VERSION),
        'uuid': kw.get('uuid', '6eb02b44-18a3-4659-8c0b-8d2802581ae4'),
        'name': kw.get('name', 'fooname'),
        'node_id': kw.get('node_id', 123),
        'node_uuid': kw.get('node_uuid',
                            '40481b96-306b-4a33-901f-795a3dc2f397'),
        'address': kw.get('address', '52:54:00:cf:2d:31'),
        'extra': kw.get('extra', {}),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
        'internal_info': kw.get('internal_info', {"bar": "buzz"}),
        'standalone_ports_supported': kw.get('standalone_ports_supported',
                                             True),
        'mode': kw.get('mode'),
        'properties': kw.get('properties', {}),
    }


def create_test_portgroup(**kw):
    """Create test portgroup entry in DB and return Portgroup DB object.

    Function to be used to create test Portgroup objects in the database.

    :param kw: kwargs with overriding values for port's attributes.
    :returns: Test Portgroup DB object.

    """
    portgroup = get_test_portgroup(**kw)
    # Let DB generate ID if it isn't specified explicitly
    for field in ('id', 'uuid'):
        if field not in kw:
            del portgroup[field]
    dbapi = db_api.get_instance()
    return dbapi.create_portgroup(portgroup)


def get_test_node_tag(**kw):
    return {
        # TODO(rloo): Replace None below with the object NodeTag VERSION,
        #             after this lands: https://review.opendev.org/#/c/233357
        'version': kw.get('version', None),
        "tag": kw.get("tag", "tag1"),
        "node_id": kw.get("node_id", "123"),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_node_tag(**kw):
    """Create test node tag entry in DB and return NodeTag DB object.

    Function to be used to create test NodeTag objects in the database.

    :param kw: kwargs with overriding values for tag's attributes.
    :returns: Test NodeTag DB object.

    """
    tag = get_test_node_tag(**kw)
    dbapi = db_api.get_instance()
    return dbapi.add_node_tag(tag['node_id'], tag['tag'])


def get_test_node_trait(**kw):
    return {
        'version': kw.get('version', trait.Trait.VERSION),
        "trait": kw.get("trait", "trait1"),
        "node_id": kw.get("node_id", "123"),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_node_trait(**kw):
    """Create test node trait entry in DB and return NodeTrait DB object.

    Function to be used to create test NodeTrait objects in the database.

    :param kw: kwargs with overriding values for trait's attributes.
    :returns: Test NodeTrait DB object.
    """
    trait = get_test_node_trait(**kw)
    dbapi = db_api.get_instance()
    return dbapi.add_node_trait(trait['node_id'], trait['trait'],
                                trait['version'])


def create_test_node_traits(traits, **kw):
    """Create test node trait entries in DB and return NodeTrait DB objects.

    Function to be used to create test NodeTrait objects in the database.

    :param traits: a list of Strings; traits to create.
    :param kw: kwargs with overriding values for trait's attributes.
    :returns: a list of test NodeTrait DB objects.
    """
    return [create_test_node_trait(trait=trait, **kw) for trait in traits]


def create_test_bios_setting(**kw):
    """Create test bios entry in DB and return BIOSSetting DB object.

    Function to be used to create test BIOSSetting object in the database.

    :param kw: kwargs with overriding values for node bios settings.
    :returns: Test BIOSSetting DB object.

    """
    bios_setting = get_test_bios_setting(**kw)
    dbapi = db_api.get_instance()
    node_id = bios_setting['node_id']
    version = bios_setting['version']
    settings = [{'name': bios_setting['name'],
                 'value': bios_setting['value'],
                 'attribute_type': bios_setting['attribute_type'],
                 'allowable_values': bios_setting['allowable_values'],
                 'read_only': bios_setting['read_only'],
                 'reset_required': bios_setting['reset_required'],
                 'unique': bios_setting['unique']}]
    return dbapi.create_bios_setting_list(node_id, settings, version)[0]


def get_test_bios_setting(**kw):
    return {
        'node_id': kw.get('node_id', '123'),
        'name': kw.get('name', 'virtualization'),
        'value': kw.get('value', 'on'),
        'attribute_type': kw.get('attribute_type', 'Enumeration'),
        'allowable_values': kw.get('allowable_values', ['on', 'off']),
        'lower_bound': kw.get('lower_bound', None),
        'max_length': kw.get('max_length', None),
        'min_length': kw.get('max_length', None),
        'read_only': kw.get('read_only', False),
        'reset_required': kw.get('reset_required', True),
        'unique': kw.get('unique', False),
        'upper_bound': kw.get('upper_bound', None),
        'version': kw.get('version', bios.BIOSSetting.VERSION),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def get_test_bios_setting_setting_list():
    return [
        {'name': 'virtualization', 'value': 'on'},
        {'name': 'hyperthread', 'value': 'enabled'},
        {'name': 'numlock', 'value': 'off'}
    ]


def get_test_allocation(**kw):
    return {
        'candidate_nodes': kw.get('candidate_nodes', []),
        'conductor_affinity': kw.get('conductor_affinity'),
        'created_at': kw.get('created_at'),
        'extra': kw.get('extra', {}),
        'id': kw.get('id', 42),
        'last_error': kw.get('last_error'),
        'name': kw.get('name'),
        'node_id': kw.get('node_id'),
        'resource_class': kw.get('resource_class', 'baremetal'),
        'state': kw.get('state', 'allocating'),
        'traits': kw.get('traits', []),
        'updated_at': kw.get('updated_at'),
        'uuid': kw.get('uuid', uuidutils.generate_uuid()),
        'version': kw.get('version', allocation.Allocation.VERSION),
        'owner': kw.get('owner', None),
    }


def create_test_allocation(**kw):
    allocation = get_test_allocation(**kw)
    if 'id' not in kw:
        del allocation['id']
    dbapi = db_api.get_instance()
    return dbapi.create_allocation(allocation)


def get_test_deploy_template(**kw):
    default_uuid = uuidutils.generate_uuid()
    return {
        'version': kw.get('version', deploy_template.DeployTemplate.VERSION),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
        'id': kw.get('id', 234),
        'name': kw.get('name', u'CUSTOM_DT1'),
        'uuid': kw.get('uuid', default_uuid),
        'steps': kw.get('steps', [get_test_deploy_template_step(
            deploy_template_id=kw.get('id', 234))]),
        'extra': kw.get('extra', {}),
    }


def get_test_deploy_template_step(**kw):
    return {
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
        'id': kw.get('id', 345),
        'deploy_template_id': kw.get('deploy_template_id', 234),
        'interface': kw.get('interface', 'raid'),
        'step': kw.get('step', 'create_configuration'),
        'args': kw.get('args', {'logical_disks': []}),
        'priority': kw.get('priority', 10),
    }


def create_test_deploy_template(**kw):
    """Create a deployment template in the DB and return DeployTemplate model.

    :param kw: kwargs with overriding values for the deploy template.
    :returns: Test DeployTemplate DB object.
    """
    template = get_test_deploy_template(**kw)
    dbapi = db_api.get_instance()
    # Let DB generate an ID if one isn't specified explicitly.
    if 'id' not in kw:
        del template['id']
    if 'steps' not in kw:
        for step in template['steps']:
            del step['id']
            del step['deploy_template_id']
    else:
        for kw_step, template_step in zip(kw['steps'], template['steps']):
            if 'id' not in kw_step:
                del template_step['id']
    return dbapi.create_deploy_template(template)


def get_test_runbook(**kw):
    default_uuid = uuidutils.generate_uuid()
    return {
        'version': kw.get('version', runbook.Runbook.VERSION),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
        'id': kw.get('id', 234),
        'name': kw.get('name', u'CUSTOM_DT1'),
        'uuid': kw.get('uuid', default_uuid),
        'steps': kw.get('steps', [get_test_runbook_step(
            runbook_id=kw.get('id', 234))]),
        'disable_ramdisk': kw.get('disable_ramdisk', False),
        'extra': kw.get('extra', {}),
        'public': kw.get('public', False),
        'owner': kw.get('owner', None),
    }


def get_test_runbook_step(**kw):
    return {
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
        'id': kw.get('id', 345),
        'runbook_id': kw.get('runbook_id', 234),
        'interface': kw.get('interface', 'raid'),
        'step': kw.get('step', 'create_configuration'),
        'args': kw.get('args', {'logical_disks': []}),
        'order': kw.get('order', 1)
    }


def create_test_runbook(**kw):
    """Create a runbook in the DB and return Runbook model.

    :param kw: kwargs with overriding values for the runbook.
    :returns: Test Runbook DB object.
    """
    runbook = get_test_runbook(**kw)
    dbapi = db_api.get_instance()
    # Let DB generate an ID if one isn't specified explicitly.
    if 'id' not in kw:
        del runbook['id']
    if 'steps' not in kw:
        for step in runbook['steps']:
            del step['id']
            del step['runbook_id']
    else:
        for kw_step, runbook_step in zip(kw['steps'], runbook['steps']):
            if 'id' not in kw_step:
                del runbook_step['id']
    return dbapi.create_runbook(runbook)


def get_test_history(**kw):
    return {
        'id': kw.get('id', 345),
        'version': kw.get('version', node_history.NodeHistory.VERSION),
        'uuid': kw.get('uuid', '6f8a5d5c-0f2d-4b2c-a62a-a38e300e3f31'),
        'node_id': kw.get('node_id', 123),
        'event': kw.get('event', 'Something is wrong'),
        'conductor': kw.get('conductor', 'host-1'),
        'severity': kw.get('severity', 'ERROR'),
        'event_type': kw.get('event_type', 'provisioning'),
        'user': kw.get('user', 'fake-user'),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_history(**kw):
    """Create test history entry in DB and return NodeHistory DB object.

    :param kw: kwargs with overriding values for port's attributes.
    :returns: Test NodeHistory DB object.
    """
    history = get_test_history(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del history['id']
    dbapi = db_api.get_instance()
    return dbapi.create_node_history(history)


def get_test_inventory(**kw):
    return {
        'id': kw.get('id', 345),
        'version': kw.get('version', node_inventory.NodeInventory.VERSION),
        'node_id': kw.get('node_id', 123),
        'inventory_data': kw.get('inventory', {"inventory": "test"}),
        'plugin_data': kw.get('plugin_data', {"pdata": {"plugin": "data"}}),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def create_test_inventory(**kw):
    """Create test inventory entry in DB and return NodeInventory DB object.

    :param kw: kwargs with overriding values for inventory attributes.
    :returns: Test NodeInventory DB object.
    """
    inventory = get_test_inventory(**kw)
    # Let DB generate ID if it isn't specified explicitly
    if 'id' not in kw:
        del inventory['id']
    dbapi = db_api.get_instance()
    return dbapi.create_node_inventory(inventory)


def create_test_firmware_component(**kw):
    """Create test Firmware Component entry in DB and return object.

    Function to be used to create test FirmwareComponent object in the DB.

    :param kw: kwargs with overriding values for firmware component
    :returns: Test FirmwareComponent DB object
    """
    fw_cmp_values = get_test_firmware_component(**kw)
    if 'id' not in kw:
        del fw_cmp_values['id']
    dbapi = db_api.get_instance()
    return dbapi.create_firmware_component(fw_cmp_values)


def get_test_firmware_component(**kw):
    # NOTE(iurygregory): update version to get from the object in the object
    # patch.
    return {
        'id': kw.get('id', 256),
        'node_id': kw.get('node_id', 123),
        'component': kw.get('component', 'bmc'),
        'initial_version': kw.get('initial_version', 'v1.0.0'),
        'current_version': kw.get('current_version', 'v1.0.0'),
        'last_version_flashed': kw.get('last_version_flashed', None),
        'version': kw.get('version', firmware.FirmwareComponent.VERSION),
        'created_at': kw.get('created_at'),
        'updated_at': kw.get('updated_at'),
    }


def get_test_firmware_component_list():
    return [
        {'component': 'bmc', 'initial_version': 'v1.0.0',
         'current_version': 'v1.0.0', 'last_version_flashed': None},
        {'component': 'BIOS', 'initial_version': 'v1.5.0',
         'current_version': 'v1.5.0', 'last_version_flashed': None},
    ]


def get_test_inspection_rule(**kw):
    default_uuid = uuidutils.generate_uuid()
    return {
        'version': kw.get('version', rule.InspectionRule.VERSION),
        'uuid': kw.get('uuid', default_uuid),
        'description': kw.get('description', 'an inspection rule'),
        'sensitive': kw.get('sensitive', False),
        'phase': kw.get('phase', 'main'),
        'priority': kw.get('priority', 0),
        'actions': kw.get('actions', [get_test_inspection_rule_action(
            inspection_rule_id=kw.get('uuid'))]),
        'conditions': kw.get('conditions', [
            get_test_inspection_rule_condition(
                inspection_rule_id=kw.get('uuid'))]),
    }


def get_test_inspection_rule_action(**kw):
    action = {
        'op': kw.get('op', 'set-attribute'),
        'args': kw.get('args', ["/driver", "idrac"]),
    }

    if 'id' in kw:
        action['id'] = kw['id']

    action.update({k: v for k, v in kw.items() if k not in action})
    return action


def get_test_inspection_rule_condition(**kw):
    condition = {
        'op': kw.get('op', 'is-true'),
        'args': kw.get('args', ["{node.auto_discovered}"]),
        'multiple': kw.get('multiple', 'any'),
    }

    if 'id' in kw:
        condition['id'] = kw['id']

    condition.update({k: v for k, v in kw.items() if k not in condition})
    return condition


def create_test_inspection_rule(**kw):
    """Create a inspection rule in the DB and return InspectionRule model.

    :param kw: kwargs with overriding values for the inspection rule.
    :returns: Test InspectionRule DB object.
    """
    inspection_rule = get_test_inspection_rule(**kw)
    dbapi = db_api.get_instance()
    if 'uuid' not in kw:
        del inspection_rule['uuid']
    return dbapi.create_inspection_rule(inspection_rule)



# Copyright 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Mapping of boot devices used when requesting the system to boot
from an alternate device.

The options presented were based on the IPMItool chassis
bootdev command. You can find the documentation at:
http://linux.die.net/man/1/ipmitool

NOTE: This module does not include all the options from ipmitool because
they don't make sense in the limited context of Ironic right now.
"""

PXE = 'pxe'
"Boot from PXE boot"

DISK = 'disk'
"Boot from default Hard-drive"

CDROM = 'cdrom'
"Boot from CD/DVD"

BIOS = 'bios'
"Boot into BIOS setup"

SAFE = 'safe'
"Boot from default Hard-drive, request Safe Mode"

WANBOOT = 'wanboot'
"Boot from Wide Area Network"

ISCSIBOOT = 'iscsiboot'
"Boot from iSCSI volume"

FLOPPY = 'floppy'
"Boot from a floppy drive"

VMEDIA_DEVICES = [DISK, CDROM, FLOPPY]
"""Devices that make sense for virtual media attachment."""

UEFIHTTP = "uefihttp"
"Boot from a UEFI HTTP(s) URL"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Internal implementation of request/response validating middleware."""

import jsonschema
from oslo_utils import timeutils
from oslo_utils import uuidutils

from ironic.common import exception
from ironic.common.i18n import _


_FORMAT_CHECKER = jsonschema.FormatChecker()


@_FORMAT_CHECKER.checks('date-time')
def _validate_datetime_format(instance: object) -> bool:
    # format checks constrain to the relevant primitive type
    # https://github.com/OAI/OpenAPI-Specification/issues/3148
    if not isinstance(instance, str):
        return True
    try:
        timeutils.parse_isotime(instance)
    except ValueError:
        return False
    else:
        return True


@_FORMAT_CHECKER.checks('uuid')
def _validate_uuid_format(instance: object) -> bool:
    # format checks constrain to the relevant primitive type
    # https://github.com/OAI/OpenAPI-Specification/issues/3148
    if not isinstance(instance, str):
        return True

    return uuidutils.is_uuid_like(instance)


class SchemaValidator:
    """A validator class

    This class is changed from Draft202012Validator to add format checkers for
    data formats that are common in the Ironic API as well as add better error
    messages.
    """

    validator = None
    validator_org = jsonschema.Draft202012Validator

    def __init__(
        self, schema, is_body=True
    ):
        self.is_body = is_body
        validator_cls = jsonschema.validators.extend(self.validator_org)
        try:
            self.validator = validator_cls(
                schema, format_checker=_FORMAT_CHECKER
            )
        except Exception:
            raise

    def validate(self, *args, **kwargs):
        try:
            self.validator.validate(*args, **kwargs)
        except jsonschema.ValidationError as e:
            error_msg = _('Schema error: %s') % e.message
            # Sometimes the root message is too generic, try to find a possible
            # root cause:
            cause = None
            current = e
            while current.context:
                current = jsonschema.exceptions.best_match(current.context)
                cause = current.message
            if cause is not None:
                error_msg += _('. Possible root cause: %s') % cause
            raise exception.InvalidParameterValue(error_msg)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from ironic.common.i18n import _


opts = [
    cfg.BoolOpt('fatal_exception_format_errors',
                default=False,
                help=_('Used if there is a formatting error when generating '
                       'an exception message (a programming error). If True, '
                       'raise an exception; if False, use the unformatted '
                       'message.'),
                deprecated_group='ironic_lib'),
]


def register_opts(conf):
    conf.register_opts(opts, group='errors')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from keystoneauth1 import exceptions as ks_exception
import openstack

from ironic.common import context
from ironic.common import exception
from ironic.conf import CONF
from ironic.drivers.modules.inspector import client
from ironic.tests.unit.db import base as db_base


@mock.patch('ironic.common.keystone.get_auth', autospec=True,
            return_value=mock.sentinel.auth)
@mock.patch('ironic.common.keystone.get_session', autospec=True,
            return_value=mock.sentinel.session)
@mock.patch.object(openstack.connection, 'Connection', autospec=True)
class GetClientTestCase(db_base.DbTestCase):

    def setUp(self):
        super(GetClientTestCase, self).setUp()
        # NOTE(pas-ha) force-reset  global inspector session object
        client._INSPECTOR_SESSION = None
        self.context = context.RequestContext(global_request_id='global')

    def test_get_client(self, mock_conn, mock_session, mock_auth):
        client.get_client(self.context)
        mock_conn.assert_called_once_with(
            session=mock.sentinel.session,
            oslo_conf=mock.ANY)
        self.assertEqual(1, mock_auth.call_count)
        self.assertEqual(1, mock_session.call_count)

    def test_get_client_standalone(self, mock_conn, mock_session, mock_auth):
        self.config(auth_strategy='noauth')
        client.get_client(self.context)
        self.assertEqual('none', CONF.inspector.auth_type)
        mock_conn.assert_called_once_with(
            session=mock.sentinel.session,
            oslo_conf=mock.ANY)
        self.assertEqual(1, mock_auth.call_count)
        self.assertEqual(1, mock_session.call_count)

    def test_get_client_connection_problem(
            self, mock_conn, mock_session, mock_auth):
        mock_conn.side_effect = ks_exception.DiscoveryFailure("")
        self.assertRaises(exception.ConfigInvalid,
                          client.get_client, self.context)
        mock_conn.assert_called_once_with(
            session=mock.sentinel.session,
            oslo_conf=mock.ANY)
        self.assertEqual(1, mock_auth.call_count)
        self.assertEqual(1, mock_session.call_count)



# Copyright (c) 2017-2019 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import uuidutils

from ironic.conductor import task_manager
from ironic.drivers.modules import agent
from ironic.drivers.modules import drac
from ironic.drivers.modules import inspector
from ironic.drivers.modules import ipxe
from ironic.drivers.modules.network import flat as flat_net
from ironic.drivers.modules import noop
from ironic.drivers.modules import redfish
from ironic.drivers.modules.storage import noop as noop_storage
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class IDRACHardwareTestCase(db_base.DbTestCase):

    def setUp(self):
        super(IDRACHardwareTestCase, self).setUp()
        self.config_temp_dir('http_root', group='deploy')
        self.config(enabled_hardware_types=['idrac'],
                    enabled_boot_interfaces=[
                        'idrac-redfish-virtual-media', 'ipxe', 'pxe'],
                    enabled_management_interfaces=['idrac-redfish'],
                    enabled_power_interfaces=['idrac-redfish', 'redfish'],
                    enabled_inspect_interfaces=[
                        'idrac-redfish', 'inspector',
                        'no-inspect'],
                    enabled_network_interfaces=['flat', 'neutron', 'noop'],
                    enabled_raid_interfaces=[
                        'idrac-redfish', 'no-raid',
                        'agent'],
                    enabled_vendor_interfaces=[
                        'idrac-redfish', 'redfish', 'no-vendor'],
                    enabled_bios_interfaces=[
                        'idrac-redfish', 'redfish', 'no-bios'])

    def _validate_interfaces(self, driver, **kwargs):
        self.assertIsInstance(
            driver.boot,
            kwargs.get('boot', ipxe.iPXEBoot))
        self.assertIsInstance(
            driver.deploy,
            kwargs.get('deploy', agent.AgentDeploy))
        self.assertIsInstance(
            driver.management,
            kwargs.get('management', drac.management.DracRedfishManagement))
        self.assertIsInstance(
            driver.power,
            kwargs.get('power', drac.power.DracRedfishPower))

        self.assertIsInstance(
            driver.bios,
            kwargs.get('bios', drac.bios.DracRedfishBIOS))

        self.assertIsInstance(
            driver.console,
            kwargs.get('console', noop.NoConsole))

        self.assertIsInstance(
            driver.inspect,
            kwargs.get('inspect', drac.inspect.DracRedfishInspect))

        self.assertIsInstance(
            driver.network,
            kwargs.get('network', flat_net.FlatNetwork))

        self.assertIsInstance(
            driver.raid,
            kwargs.get('raid', drac.raid.DracRedfishRAID))

        self.assertIsInstance(
            driver.storage,
            kwargs.get('storage', noop_storage.NoopStorage))

        self.assertIsInstance(
            driver.vendor,
            kwargs.get('vendor',
                       drac.vendor_passthru.DracRedfishVendorPassthru))

    def test_default_interfaces(self):
        node = obj_utils.create_test_node(self.context, driver='idrac')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task.driver)

    def test_override_with_inspector(self):
        node = obj_utils.create_test_node(self.context, driver='idrac',
                                          inspect_interface='inspector')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task.driver,
                                      inspect=inspector.Inspector)

    def test_override_with_raid(self):
        for iface, impl in [('agent', agent.AgentRAID),
                            ('no-raid', noop.NoRAID)]:
            node = obj_utils.create_test_node(self.context,
                                              uuid=uuidutils.generate_uuid(),
                                              driver='idrac',
                                              raid_interface=iface)
            with task_manager.acquire(self.context, node.id) as task:
                self._validate_interfaces(task.driver, raid=impl)

    def test_override_with_redfish_raid(self):
        node = obj_utils.create_test_node(self.context,
                                          uuid=uuidutils.generate_uuid(),
                                          driver='idrac',
                                          raid_interface='idrac-redfish')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(task.driver,
                                      raid=drac.raid.DracRedfishRAID)

    def test_override_with_redfish_vendor(self):
        for iface, impl in [('redfish',
                             redfish.vendor.RedfishVendorPassthru),
                            ('no-vendor', noop.NoVendor)]:
            node = obj_utils.create_test_node(self.context,
                                              uuid=uuidutils.generate_uuid(),
                                              driver='idrac',
                                              vendor_interface=iface)
            with task_manager.acquire(self.context, node.id) as task:
                self._validate_interfaces(task.driver,
                                          vendor=impl)

    def test_override_with_redfish_management_and_power(self):
        node = obj_utils.create_test_node(self.context, driver='idrac',
                                          management_interface='idrac-redfish',
                                          power_interface='redfish')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(
                task.driver,
                management=drac.management.DracRedfishManagement,
                power=redfish.power.RedfishPower)

    def test_override_with_redfish_bios(self):
        for iface, impl in [('redfish', redfish.bios.RedfishBIOS),
                            ('no-bios', noop.NoBIOS)]:
            node = obj_utils.create_test_node(self.context,
                                              uuid=uuidutils.generate_uuid(),
                                              driver='idrac',
                                              bios_interface=iface)
            with task_manager.acquire(self.context, node.id) as task:
                self._validate_interfaces(task.driver, bios=impl)

    def test_override_with_redfish_inspect(self):
        node = obj_utils.create_test_node(self.context, driver='idrac',
                                          inspect_interface='idrac-redfish')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(
                task.driver,
                inspect=drac.inspect.DracRedfishInspect)

    def test_override_with_redfish_virtual_media_boot(self):
        node = obj_utils.create_test_node(
            self.context, driver='idrac',
            boot_interface='idrac-redfish-virtual-media')
        with task_manager.acquire(self.context, node.id) as task:
            self._validate_interfaces(
                task.driver,
                boot=drac.boot.DracRedfishVirtualMediaBoot)



# Copyright 2016 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
from urllib import parse as urlparse

from oslo_config import cfg
from oslo_log import log
from oslo_utils import uuidutils
from pecan import rest

from ironic import api
from ironic.api.controllers.v1 import node as node_ctl
from ironic.api.controllers.v1 import notification_utils as notify
from ironic.api.controllers.v1 import utils as api_utils
from ironic.api.controllers.v1 import versions
from ironic.api import method
from ironic.common import args
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import states
from ironic.common import utils
from ironic.drivers.modules import inspect_utils
from ironic import objects


CONF = cfg.CONF
LOG = log.getLogger(__name__)

_LOOKUP_RETURN_FIELDS = ['uuid', 'properties', 'instance_info',
                         'driver_internal_info']
AGENT_VALID_STATES = ['start', 'end', 'error']


def config(token):
    return {
        'metrics': {
            'backend': CONF.metrics.agent_backend,
            'prepend_host': CONF.metrics.agent_prepend_host,
            'prepend_uuid': CONF.metrics.agent_prepend_uuid,
            'prepend_host_reverse': CONF.metrics.agent_prepend_host_reverse,
            'global_prefix': CONF.metrics.agent_global_prefix
        },
        'metrics_statsd': {
            'statsd_host': CONF.metrics_statsd.agent_statsd_host,
            'statsd_port': CONF.metrics_statsd.agent_statsd_port
        },
        'heartbeat_timeout': CONF.api.ramdisk_heartbeat_timeout,
        'agent_token': token,
        # Since this is for the Victoria release, we send this as an
        # explicit True statement for newer agents to lock the setting
        # and behavior into place.
        'agent_token_required': True,
        'agent_md5_checksum_enable': CONF.agent.allow_md5_checksum,
        'disable_deep_image_inspection': CONF.conductor.disable_deep_image_inspection,  # noqa
        'permitted_image_formats': CONF.conductor.permitted_image_formats,
    }


def convert_with_links(node):
    token = node.driver_internal_info.get('agent_secret_token')
    node = node_ctl.node_convert_with_links(node, _LOOKUP_RETURN_FIELDS)
    return {'node': node, 'config': config(token)}


def get_valid_mac_addresses(addresses, node_uuid=None):
    if addresses is None:
        addresses = []

    valid_addresses = []
    invalid_addresses = []
    for addr in addresses:
        try:
            mac = utils.validate_and_normalize_mac(addr)
            valid_addresses.append(mac)
        except exception.InvalidMAC:
            invalid_addresses.append(addr)

    if invalid_addresses:
        node_log = ('' if not node_uuid
                    else '(Node UUID: %s)' % node_uuid)
        LOG.warning('The following MAC addresses "%(addrs)s" are '
                    'invalid and will be ignored by the lookup '
                    'request %(node)s',
                    {'addrs': ', '.join(invalid_addresses),
                     'node': node_log})

    return valid_addresses


class LookupController(rest.RestController):
    """Controller handling node lookup for a deploy ramdisk."""

    def lookup_allowed(self, node):
        if utils.fast_track_enabled(node):
            return (
                node.provision_state in states.FASTTRACK_LOOKUP_ALLOWED_STATES
            )
        else:
            return node.provision_state in states.LOOKUP_ALLOWED_STATES

    @method.expose()
    @args.validate(addresses=args.string_list, node_uuid=args.uuid)
    def get_all(self, addresses=None, node_uuid=None):
        """Look up a node by its MAC addresses and optionally UUID.

        If the "restrict_lookup" option is set to True (the default), limit
        the search to nodes in certain transient states (e.g. deploy wait).

        :param addresses: list of MAC addresses for a node.
        :param node_uuid: UUID of a node.
        :raises: NotFound if requested API version does not allow this
            endpoint.
        :raises: NotFound if suitable node was not found or node's provision
            state is not allowed for the lookup.
        :raises: IncompleteLookup if neither node UUID nor any valid MAC
            address was provided.
        """
        if not api_utils.allow_ramdisk_endpoints():
            raise exception.NotFound()

        api_utils.check_policy('baremetal:driver:ipa_lookup')

        # Validate the list of MAC addresses
        valid_addresses = get_valid_mac_addresses(addresses)
        if not valid_addresses and not node_uuid:
            raise exception.IncompleteLookup()

        try:
            if node_uuid:
                node = objects.Node.get_by_uuid(
                    api.request.context, node_uuid)
            else:
                node = objects.Node.get_by_port_addresses(
                    api.request.context, valid_addresses)
        except exception.NotFound as e:
            # NOTE(dtantsur): we are reraising the same exception to make sure
            # we don't disclose the difference between nodes that are not found
            # at all and nodes in a wrong state by different error messages.
            LOG.error('No node has been found during lookup: %s', e)
            raise exception.NotFound()

        if CONF.api.restrict_lookup and not self.lookup_allowed(node):
            LOG.error('Lookup is not allowed for node %(node)s in the '
                      'provision state %(state)s',
                      {'node': node.uuid, 'state': node.provision_state})
            raise exception.NotFound()

        if api_utils.allow_agent_token():
            try:
                topic = api.request.rpcapi.get_topic_for(node)
            except exception.NoValidHost as e:
                e.code = http_client.BAD_REQUEST
                raise

            found_node = api.request.rpcapi.get_node_with_token(
                api.request.context, node.uuid, topic=topic)
        else:
            found_node = node
        return convert_with_links(found_node)


class HeartbeatController(rest.RestController):
    """Controller handling heartbeats from deploy ramdisk."""

    @method.expose(status_code=http_client.ACCEPTED)
    @args.validate(node_ident=args.uuid_or_name, callback_url=args.string,
                   agent_version=args.string, agent_token=args.string,
                   agent_verify_ca=args.string, agent_status=args.string,
                   agent_status_message=args.string)
    def post(self, node_ident, callback_url, agent_version=None,
             agent_token=None, agent_verify_ca=None, agent_status=None,
             agent_status_message=None):
        """Process a heartbeat from the deploy ramdisk.

        :param node_ident: the UUID or logical name of a node.
        :param callback_url: the URL to reach back to the ramdisk.
        :param agent_version: The version of the agent that is heartbeating.
            ``None`` indicates that the agent that is heartbeating is a version
            before sending agent_version was introduced so agent v3.0.0 (the
            last release before sending agent_version was introduced) will be
            assumed.
        :param agent_token: randomly generated validation token.
        :param agent_verify_ca: TLS certificate to use to connect to the agent.
        :param agent_status: Current status of the heartbeating agent. Used by
            anaconda ramdisk to send status back to Ironic. The valid states
            are 'start', 'end', 'error'
        :param agent_status_message: Optional status message describing current
            agent_status
        :raises: NodeNotFound if node with provided UUID or name was not found.
        :raises: InvalidUuidOrName if node_ident is not valid name or UUID.
        :raises: NoValidHost if RPC topic for node could not be retrieved.
        :raises: NotFound if requested API version does not allow this
            endpoint.
        """
        if not api_utils.allow_ramdisk_endpoints():
            raise exception.NotFound()

        if agent_version and not api_utils.allow_agent_version_in_heartbeat():
            raise exception.InvalidParameterValue(
                _('Field "agent_version" not recognised'))

        if ((agent_status or agent_status_message)
                and not api_utils.allow_status_in_heartbeat()):
            raise exception.InvalidParameterValue(
                _('Fields "agent_status" and "agent_status_message" '
                  'not recognised.')
            )

        api_utils.check_policy('baremetal:node:ipa_heartbeat')

        if (agent_verify_ca is not None
                and not api_utils.allow_verify_ca_in_heartbeat()):
            raise exception.InvalidParameterValue(
                _('Field "agent_verify_ca" not recognised in this version'))

        rpc_node = api_utils.get_rpc_node_with_suffix(node_ident)
        dii = rpc_node['driver_internal_info']
        agent_url = dii.get('agent_url')
        try:
            # NOTE(TheJulia): Use of urllib.urlparse is not a security
            # guard, but detecting oddities and incorrect formatting
            # https://docs.python.org/3/library/urllib.parse.html#url-parsing-security  # noqa
            parsed_url = urlparse.urlparse(callback_url)
            # Check if http (compatibility), or https (much newer agents)
            if 'http' not in parsed_url.scheme:
                raise ValueError
            callback_url = parsed_url.geturl()
        except ValueError:
            if callback_url != "":
                # Anaconda deploy interface sends a empty callback url, since
                # it is a one way heartbeat.
                raise exception.InvalidParameterValue(
                    _('An issue with the supplied "callback_url" has been '
                      'detected.'))
        # If we have an agent_url on file, and we get something different
        # we should fail because this is unexpected behavior of the agent.
        if agent_url is not None and agent_url != callback_url:
            LOG.error('Received heartbeat for node %(node)s with '
                      'callback URL %(url)s. This is not expected, '
                      'and the heartbeat will not be processed.',
                      {'node': rpc_node.uuid, 'url': callback_url})
            raise exception.Invalid(
                _('Detected change in ramdisk provided '
                  '"callback_url"'))
        # NOTE(TheJulia): If tokens are required, lets go ahead and fail the
        # heartbeat very early on.
        if agent_token is None:
            LOG.error('Agent heartbeat received for node %(node)s '
                      'without an agent token.', {'node': node_ident})
            raise exception.InvalidParameterValue(
                _('Agent token is required for heartbeat processing.'))

        if agent_status is not None and agent_status not in AGENT_VALID_STATES:
            valid_states = ','.join(AGENT_VALID_STATES)
            LOG.error('Agent heartbeat received for node %(node)s '
                      'has an invalid agent status: %(agent_status)s. '
                      'Valid states are %(valid_states)s ',
                      {'node': node_ident, 'agent_status': agent_status,
                       'valid_states': valid_states})
            msg = (_('Agent status is invalid. Valid states are %s.') %
                   valid_states)
            raise exception.InvalidParameterValue(msg)

        try:
            topic = api.request.rpcapi.get_topic_for(rpc_node)
        except exception.NoValidHost as e:
            e.code = http_client.BAD_REQUEST
            raise

        api.request.rpcapi.heartbeat(
            api.request.context, rpc_node.uuid, callback_url,
            agent_version, agent_token, agent_verify_ca, agent_status,
            agent_status_message, topic=topic)


DATA_VALIDATOR = args.schema({
    'type': 'object',
    'properties': {
        # This validator defines a minimal acceptable inventory.
        'inventory': {
            'type': 'object',
            'properties': {
                'bmc_address': {'type': ['string', 'null']},
                'bmc_v6address': {'type': ['string', 'null']},
                'interfaces': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'mac_address': {'type': 'string'},
                        },
                        'required': ['mac_address'],
                        'additionalProperties': True,
                    },
                    'minItems': 1,
                },
            },
            'required': ['interfaces'],
            'additionalProperties': True,
        },
    },
    'required': ['inventory'],
    'additionalProperties': True,
})


class ContinueInspectionController(rest.RestController):
    """Controller handling inspection data from deploy ramdisk."""

    def _auto_enroll(self, macs, bmc_addresses):
        context = api.request.context
        new_node = objects.Node(
            context,
            conductor_group='',  # TODO(dtantsur): default_conductor_group
            driver=CONF.auto_discovery.driver,
            provision_state=states.ENROLL,
            resource_class=CONF.default_resource_class,
            uuid=uuidutils.generate_uuid())

        try:
            topic = api.request.rpcapi.get_topic_for(new_node)
        except exception.NoValidHost as e:
            LOG.error("Failed to find a conductor to handle the newly "
                      "enrolled node with driver %s: %s", new_node.driver, e)
            # NOTE(dtantsur): do not disclose any information to the caller
            raise exception.IronicException()

        LOG.info("Enrolling the newly discovered node %(uuid)s with driver "
                 "%(driver)s, MAC addresses [%(macs)s] and BMC address(es) "
                 "[%(bmc)s]",
                 {'driver': new_node.driver,
                  'uuid': new_node.uuid,
                  'macs': ', '.join(macs or ()),
                  'bmc': ', '.join(bmc_addresses or ())})

        notify.emit_start_notification(context, new_node, 'create')
        with notify.handle_error_notification(context, new_node, 'create'):
            try:
                node = api.request.rpcapi.create_node(
                    context, new_node, topic=topic)
            except exception.IronicException:
                LOG.exception("Failed to enroll node with driver %s",
                              new_node.driver)
                # NOTE(dtantsur): do not disclose any information to the caller
                raise exception.IronicException()

        return node, topic

    @method.expose(status_code=http_client.ACCEPTED)
    @method.body('data')
    @args.validate(data=DATA_VALIDATOR, node_uuid=args.uuid)
    def post(self, data, node_uuid=None):
        """Process a introspection data from the deploy ramdisk.

        :param data: Introspection data.
        :param node_uuid: UUID of a node.
        :raises: InvalidParameterValue if node_uuid is a valid UUID.
        :raises: NoValidHost if RPC topic for node could not be retrieved.
        :raises: NotFound if requested API version does not allow this
            endpoint or if lookup fails.
        """
        if (not api_utils.allow_continue_inspection_endpoint()
                # Node UUID support is a new addition
                or (node_uuid
                    and not api_utils.new_continue_inspection_endpoint())):
            raise exception.NotFound(
                # This is a small lie: 1.1 is accepted as well, but no need
                # to really advertise this fact, it's only for compatibility.
                _('API version 1.%d or newer is required')
                % versions.MINOR_84_CONTINUE_INSPECTION)

        api_utils.check_policy('baremetal:driver:ipa_continue_inspection')

        inventory = data.pop('inventory')
        macs = get_valid_mac_addresses(
            iface['mac_address'] for iface in inventory['interfaces'])
        bmc_addresses = list(
            filter(None, (inventory.get('bmc_address'),
                          inventory.get('bmc_v6address')))
        )
        if not macs and not bmc_addresses and not node_uuid:
            raise exception.BadRequest(_('No lookup information provided'))

        try:
            rpc_node = inspect_utils.lookup_node(
                api.request.context, macs, bmc_addresses, node_uuid=node_uuid)
        except inspect_utils.AutoEnrollPossible:
            if not CONF.auto_discovery.enabled:
                raise exception.NotFound()
            rpc_node, topic = self._auto_enroll(macs, bmc_addresses)
            # TODO(dtantsur): consider adding a Node-level property to make
            # newly discovered nodes searchable via API. The flag in
            # plugin_data is for compatibility with ironic-inspector.
            data[inspect_utils.AUTO_DISCOVERED_FLAG] = True
        else:
            try:
                topic = api.request.rpcapi.get_topic_for(rpc_node)
            except exception.NoValidHost as e:
                e.code = http_client.BAD_REQUEST
                raise

        if api_utils.new_continue_inspection_endpoint():
            # This has to happen before continue_inspection since processing
            # the data may take significant time, and creating a token required
            # a lock on the node.
            rpc_node = api.request.rpcapi.get_node_with_token(
                api.request.context, rpc_node.uuid, topic=topic)

        api.request.rpcapi.continue_inspection(
            api.request.context, rpc_node.uuid, inventory=inventory,
            plugin_data=data, topic=topic)

        if api_utils.new_continue_inspection_endpoint():
            return convert_with_links(rpc_node)
        else:
            # Compatibility with ironic-inspector
            return {'uuid': rpc_node.uuid}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fcntl
import os
import time

from oslo_log import log

from ironic.conf import CONF

LOG = log.getLogger(__name__)


def update(allow_macs, deny_macs, allow_unknown=None):
    """Update only the given MACs.

    MACs not in either lists are ignored.

    :param allow_macs: MACs to allow in dnsmasq.
    :param deny_macs: MACs to disallow in dnsmasq.
    :param allow_unknown: If set to True, unknown MACs are also allowed.
        Setting it to False does nothing in this call.
    """
    for mac in allow_macs:
        _add_mac_to_allowlist(mac)
    for mac in deny_macs:
        _add_mac_to_denylist(mac)
    if allow_unknown:
        _configure_unknown_hosts(True)


def sync(allow_macs, deny_macs, allow_unknown):
    """Conduct a complete sync of the state.

    Unlike ``update``, MACs not in either list are handled according
    to ``allow_unknown``.

    :param allow_macs: MACs to allow in dnsmasq.
    :param deny_macs: MACs to disallow in dnsmasq.
    :param allow_unknown: Whether to allow access to dnsmasq to unknown
        MACs.
    """
    allow_macs = set(allow_macs)
    deny_macs = set(deny_macs)

    known_macs = allow_macs.union(deny_macs)
    current_denylist, current_allowlist = _get_deny_allow_lists()
    removed_macs = current_denylist.union(current_allowlist).difference(
        known_macs)

    update(allow_macs=allow_macs.difference(current_allowlist),
           deny_macs=deny_macs.difference(current_denylist))

    # Allow or deny unknown hosts and MACs not kept in ironic
    # NOTE(hjensas): Treat unknown hosts and MACs not kept in ironic the
    # same. Neither should boot the inspection image unless inspection
    # is active. Deleted MACs must be added to the allow list when
    # inspection is active in case the host is re-enrolled.
    _configure_unknown_hosts(allow_unknown)
    _configure_removedlist(removed_macs, allow_unknown)


_EXCLUSIVE_WRITE_ATTEMPTS = 10
_EXCLUSIVE_WRITE_ATTEMPTS_DELAY = 0.01

_MAC_DENY_LEN = len('ff:ff:ff:ff:ff:ff,ignore\n')
_MAC_ALLOW_LEN = len('ff:ff:ff:ff:ff:ff\n')
_UNKNOWN_HOSTS_FILE = 'unknown_hosts_filter'
_DENY_UNKNOWN_HOSTS = '*:*:*:*:*:*,ignore\n'
_ALLOW_UNKNOWN_HOSTS = '*:*:*:*:*:*\n'


def _get_deny_allow_lists():
    """Get addresses currently denied by dnsmasq.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid.
    :returns: tuple with 2 elements: a set of MACs currently denied by dnsmasq
        and a set of allowed MACs.
    """
    hostsdir = CONF.pxe_filter.dhcp_hostsdir
    # MACs in the allow list lack the ,ignore directive
    denylist = set()
    allowlist = set()
    for mac in os.listdir(hostsdir):
        if os.stat(os.path.join(hostsdir, mac)).st_size == _MAC_DENY_LEN:
            denylist.add(mac)
        if os.stat(os.path.join(hostsdir, mac)).st_size == _MAC_ALLOW_LEN:
            allowlist.add(mac)

    return denylist, allowlist


def _exclusive_write_or_pass(path, buf):
    """Write exclusively or pass if path locked.

    The intention is to be able to run multiple instances of the filter on the
    same node in multiple inspector processes.

    :param path: where to write to
    :param buf: the content to write
    :raises: FileNotFoundError, IOError
    :returns: True if the write was successful.
    """
    # NOTE(milan) line-buffering enforced to ensure dnsmasq record update
    # through inotify, which reacts on f.close()
    with open(path, 'w', 1) as f:
        for attempt in range(_EXCLUSIVE_WRITE_ATTEMPTS):
            try:
                fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
                f.write(buf)
                # Go ahead and flush the data now instead of waiting until
                # after the automatic flush with the file close after the
                # file lock is released.
                f.flush()
                return True
            except BlockingIOError:
                LOG.debug('%s locked; will try again (later)', path)
                time.sleep(_EXCLUSIVE_WRITE_ATTEMPTS_DELAY)
                continue
            finally:
                fcntl.flock(f, fcntl.LOCK_UN)
    LOG.debug('Failed to write the exclusively-locked path: %(path)s for '
              '%(attempts)s times', {'attempts': _EXCLUSIVE_WRITE_ATTEMPTS,
                                     'path': path})
    return False


def _configure_removedlist(macs, allowed):
    """Manages a dhcp_hostsdir allow/deny record for removed macs

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
    :returns: None.
    """
    hostsdir = CONF.pxe_filter.dhcp_hostsdir

    for mac in macs:
        file_size = os.stat(os.path.join(hostsdir, mac)).st_size
        if allowed:
            if file_size != _MAC_ALLOW_LEN:
                _add_mac_to_allowlist(mac)
        else:
            if file_size != _MAC_DENY_LEN:
                _add_mac_to_denylist(mac)


def _configure_unknown_hosts(enabled):
    """Manages a dhcp_hostsdir allow/deny record for unknown macs.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
             IOError in case the dhcp host unknown file isn't writable.
    :returns: None.
    """
    path = os.path.join(CONF.pxe_filter.dhcp_hostsdir, _UNKNOWN_HOSTS_FILE)

    if enabled:
        wildcard_filter = _ALLOW_UNKNOWN_HOSTS
        log_wildcard_filter = 'allow'
    else:
        wildcard_filter = _DENY_UNKNOWN_HOSTS
        log_wildcard_filter = 'deny'

    # Don't update if unknown hosts are already in the deny/allow-list
    try:
        if os.stat(path).st_size == len(wildcard_filter):
            return
    except FileNotFoundError:
        pass

    if _exclusive_write_or_pass(path, '%s' % wildcard_filter):
        LOG.debug('A %s record for all unknown hosts using wildcard mac '
                  'created', log_wildcard_filter)
    else:
        LOG.warning('Failed to %s unknown hosts using wildcard mac; '
                    'retrying next periodic sync time', log_wildcard_filter)


def _add_mac_to_denylist(mac):
    """Creates a dhcp_hostsdir deny record for the MAC.

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
             IOError in case the dhcp host MAC file isn't writable.
    :returns: None.
    """
    path = os.path.join(CONF.pxe_filter.dhcp_hostsdir, mac)
    if _exclusive_write_or_pass(path, '%s,ignore\n' % mac):
        LOG.debug('MAC %s added to the deny list', mac)
    else:
        LOG.warning('Failed to add MAC %s to the deny list; retrying next '
                    'periodic sync time', mac)


def _add_mac_to_allowlist(mac):
    """Update the dhcp_hostsdir record for the MAC adding it to allow list

    :raises: FileNotFoundError in case the dhcp_hostsdir is invalid,
             IOError in case the dhcp host MAC file isn't writable.
    :returns: None.
    """
    path = os.path.join(CONF.pxe_filter.dhcp_hostsdir, mac)
    # remove the ,ignore directive
    if _exclusive_write_or_pass(path, '%s\n' % mac):
        LOG.debug('MAC %s removed from the deny list', mac)
    else:
        LOG.warning('Failed to remove MAC %s from the deny list; retrying '
                    'next periodic sync time', mac)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""LLDP Processing Hook for basic TLVs"""

import binascii

from oslo_log import log as logging

from ironic.drivers.modules.inspector.hooks import base
from ironic.drivers.modules.inspector import lldp_parsers


LOG = logging.getLogger(__name__)


class ParseLLDPHook(base.InspectionHook):
    """Process LLDP packet fields and store them in plugin_data['parsed_lldp']

    Convert binary LLDP information into a readable form. Loop through raw
    LLDP TLVs and parse those from the basic management, 802.1, and 802.3 TLV
    sets. Store parsed data in the plugin_data as a new parsed_lldp dictionary
    with interface names as keys.
    """

    def _parse_lldp_tlvs(self, tlvs, node_uuid):
        """Parse LLDP TLVs into a dictionary of name/value pairs

        :param tlvs: List of raw TLVs
        :param node_uuid: UUID of the node being inspected
        :returns: Dictionary of name/value pairs. The LLDP user-friendly
                  names, e.g. "switch_port_id" are the keys.
        """
        # Generate name/value pairs for each TLV supported by this plugin.
        parser = lldp_parsers.LLDPBasicMgmtParser(node_uuid)

        for tlv_type, tlv_value in tlvs:
            try:
                data = bytearray(binascii.a2b_hex(tlv_value))
            except TypeError as e:
                LOG.warning(
                    'TLV value for TLV type %(tlv_type)d is not in correct '
                    'format, value must be in hexadecimal: %(msg)s. Node: '
                    '%(node)s', {'tlv_type': tlv_type, 'msg': e,
                                 'node': node_uuid})
                continue

            try:
                parsed_tlv = parser.parse_tlv(tlv_type, data)
            except UnicodeDecodeError as e:
                LOG.warning("LLDP TLV type %(tlv_type)d from Node '%(node)s' "
                            "can't be decoded: %(exc)s",
                            {'tlv_type': tlv_type, 'exc': e,
                             'node': node_uuid})
                continue

            if parsed_tlv:
                LOG.debug("Handled TLV type %d. Node: %s", tlv_type, node_uuid)
            else:
                LOG.debug("LLDP TLV type %d not handled. Node: %s", tlv_type,
                          node_uuid)
        return parser.nv_dict

    def __call__(self, task, inventory, plugin_data):
        """Process LLDP data and update plugin_data with processed data"""

        lldp_raw = plugin_data.get('lldp_raw') or {}

        for interface in inventory['interfaces']:
            if_name = interface['name']
            tlvs = lldp_raw.get(if_name) or interface.get('lldp')
            if tlvs is None:
                LOG.warning("No LLDP Data found for interface %s of node %s",
                            if_name, task.node.uuid)
                continue

            LOG.debug("Processing LLDP Data for interface %s of node %s",
                      if_name, task.node.uuid)

            # Store LLDP data per interface in plugin_data[parsed_lldp]
            nv = self._parse_lldp_tlvs(tlvs, task.node.uuid)
            if nv:
                if plugin_data.get('parsed_lldp'):
                    plugin_data['parsed_lldp'].update({if_name: nv})
                else:
                    plugin_data['parsed_lldp'] = {if_name: nv}



#
# Copyright 2015 Rackspace, Inc
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
from http import client as http_client
import json
import sys
import traceback

from oslo_config import cfg
from oslo_log import log
import pecan

LOG = log.getLogger(__name__)


pecan_json_decorate = pecan.expose(
    content_type='application/json',
    generic=False)


def expose(status_code=None):

    def decorate(f):

        @functools.wraps(f)
        def callfunction(self, *args, **kwargs):
            try:
                result = f(self, *args, **kwargs)
                if status_code:
                    pecan.response.status = status_code

            except Exception:
                try:
                    exception_info = sys.exc_info()
                    orig_exception = exception_info[1]
                    orig_code = getattr(orig_exception, 'code', None)
                    result = format_exception(
                        exception_info,
                        cfg.CONF.debug_tracebacks_in_api
                    )
                finally:
                    del exception_info

                if orig_code and orig_code in http_client.responses:
                    pecan.response.status = orig_code
                else:
                    pecan.response.status = 500

            def _empty():
                # This is for a pecan workaround originally in WSME,
                # but the original issue description is in an issue tracker
                # that is now offline
                pecan.request.pecan['content_type'] = None
                pecan.response.content_type = None

            # never return content for NO_CONTENT
            if pecan.response.status_code == 204:
                return _empty()

            # don't encode None for ACCEPTED responses
            if result is None and pecan.response.status_code == 202:
                return _empty()

            return json.dumps(result)

        pecan_json_decorate(callfunction)
        return callfunction

    return decorate


def body(body_arg):
    """Decorator which places HTTP request body JSON into a method argument

    :param body_arg: Name of argument to populate with body JSON
    """

    def inner_function(function):

        @functools.wraps(function)
        def inner_body(*args, **kwargs):

            if pecan.request.body:
                data = pecan.request.json
            else:
                data = {}
            if isinstance(data, dict):
                # remove any keyword arguments which pecan has
                # extracted from the body
                for field in data.keys():
                    kwargs.pop(field, None)

            kwargs[body_arg] = data

            return function(*args, **kwargs)
        return inner_body
    return inner_function


def format_exception(excinfo, debug=False):
    """Extract information that can be sent to the client."""
    error = excinfo[1]
    code = getattr(error, 'code', None)
    if code and code in http_client.responses and (400 <= code < 500):
        faultstring = (error.faultstring if hasattr(error, 'faultstring')
                       else str(error))
        faultcode = getattr(error, 'faultcode', 'Client')
        r = dict(faultcode=faultcode,
                 faultstring=faultstring)
        LOG.debug("Client-side error: %s", r['faultstring'])
        r['debuginfo'] = None
        return r
    else:
        faultstring = str(error)
        debuginfo = "\n".join(traceback.format_exception(*excinfo))

        LOG.error('Server-side error: "%s". Detail: \n%s',
                  faultstring, debuginfo)

        faultcode = getattr(error, 'faultcode', 'Server')
        r = dict(faultcode=faultcode, faultstring=faultstring)
        if debug:
            r['debuginfo'] = debuginfo
        else:
            r['debuginfo'] = None
        return r



# Copyright 2016 Hewlett Packard Enterprise Development Company LP.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common.i18n import _
from ironic.conf import auth

opts = [
    cfg.IntOpt('retries',
               default=3,
               deprecated_for_removal=True,
               deprecated_reason=_('Replaced by status_code_retries and '
                                   'status_code_retry_delay.'),
               help=_('DEPRECATED: Client retries in the case of a failed '
                      'request.')),
    cfg.IntOpt('action_retries',
               default=3,
               help=_('Number of retries in the case of a failed '
                      'action (currently only used when detaching '
                      'volumes).')),
    cfg.IntOpt('action_retry_interval',
               default=5,
               help=_('Retry interval in seconds in the case of a failed '
                      'action (only specific actions are retried).')),
]


def register_opts(conf):
    conf.register_opts(opts, group='cinder')
    auth.register_auth_opts(conf, 'cinder', service_type='volumev3')


def list_opts():
    return auth.add_auth_opts(opts, service_type='volumev3')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import random
import string
from unittest import mock

from oslo_utils import uuidutils

from ironic.common import states
from ironic.conf import CONF
from ironic.pxe_filter import dnsmasq
from ironic.pxe_filter import service as pxe_filter_service
from ironic.tests.unit.db import base as test_base
from ironic.tests.unit.db import utils as db_utils


def generate_mac():
    return ':'.join(''.join(random.choice(string.hexdigits) for _ in range(2))
                    for _ in range(6))


@mock.patch.object(dnsmasq, 'sync', autospec=True)
class TestSync(test_base.DbTestCase):

    def setUp(self):
        super().setUp()
        self.service = pxe_filter_service.PXEFilterManager('host')

    def test_no_nodes(self, mock_sync):
        self.service._sync(self.dbapi)
        mock_sync.assert_called_once_with([], [], False)

    def test_no_nodes_with_discovery(self, mock_sync):
        CONF.set_override('enabled', True, group='auto_discovery')
        self.service._sync(self.dbapi)
        mock_sync.assert_called_once_with([], [], True)

    def test_sync(self, mock_sync):
        on_inspection = [
            db_utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                      provision_state=state,
                                      inspect_interface='agent')
            for state in (states.INSPECTWAIT, states.INSPECTING)
        ]
        not_on_inspection = [
            db_utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                      provision_state=state,
                                      inspect_interface='agent')
            for state in (states.ACTIVE, states.AVAILABLE, states.INSPECTFAIL)
        ]
        ignored = db_utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                            provision_state=states.INSPECTING,
                                            inspect_interface='no-inspect')
        ignored_port = db_utils.create_test_port(
            uuid=uuidutils.generate_uuid(),
            node_id=ignored.id,
            address=generate_mac())

        allow_macs, deny_macs = set(), {ignored_port.address}
        for count, node in enumerate(on_inspection):
            for _i in range(count):
                port = db_utils.create_test_port(
                    uuid=uuidutils.generate_uuid(),
                    node_id=node.id,
                    address=generate_mac())
                allow_macs.add(port.address)
        for count, node in enumerate(not_on_inspection):
            for _i in range(count):
                port = db_utils.create_test_port(
                    uuid=uuidutils.generate_uuid(),
                    node_id=node.id,
                    address=generate_mac())
                deny_macs.add(port.address)

        self.service._sync(self.dbapi)
        mock_sync.assert_called_once_with(mock.ANY, mock.ANY, True)
        self.assertEqual(allow_macs, set(mock_sync.call_args.args[0]))
        self.assertEqual(deny_macs, set(mock_sync.call_args.args[1]))

    def test_nothing_on_inspection(self, mock_sync):
        not_on_inspection = [
            db_utils.create_test_node(uuid=uuidutils.generate_uuid(),
                                      provision_state=state,
                                      inspect_interface='agent')
            for state in (states.ACTIVE, states.AVAILABLE, states.INSPECTFAIL)
        ]

        deny_macs = set()
        for count, node in enumerate(not_on_inspection):
            for _i in range(count):
                port = db_utils.create_test_port(
                    uuid=uuidutils.generate_uuid(),
                    node_id=node.id,
                    address=generate_mac())
                deny_macs.add(port.address)

        self.service._sync(self.dbapi)
        mock_sync.assert_called_once_with([], mock.ANY, False)
        self.assertEqual(deny_macs, set(mock_sync.call_args.args[1]))


class TestManager(test_base.DbTestCase):

    @mock.patch('eventlet.spawn_after', lambda delay, func: func())
    @mock.patch('eventlet.event.Event', autospec=True)
    @mock.patch.object(pxe_filter_service.PXEFilterManager, '_sync',
                       autospec=True)
    def test_init_and_run(self, mock_sync, mock_event):
        mock_wait = mock_event.return_value.wait
        mock_wait.side_effect = [None, None, True]
        mock_sync.side_effect = [None, RuntimeError(), None]

        service = pxe_filter_service.PXEFilterManager('example.com')
        service.init_host(mock.sentinel.context)

        mock_sync.assert_called_with(service, mock.ANY)
        self.assertEqual(3, mock_sync.call_count)
        mock_wait.assert_called_with(timeout=45)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import os.path
import shutil
from urllib import parse as urlparse

from oslo_log import log

from ironic.common import exception
from ironic.common import swift
from ironic.common import utils
from ironic.conf import CONF

LOG = log.getLogger(__name__)


class AbstractPublisher(metaclass=abc.ABCMeta):
    """Abstract base class for publishing images via HTTP."""

    @abc.abstractmethod
    def publish(self, source_path, file_name=None):
        """Publish an image.

        :param source_path: Path to the source file.
        :param file_name: Destination file name. If None, the file component
            of source_path is used.
        :return: The HTTP URL of the published image.
        """

    @abc.abstractmethod
    def unpublish(self, file_name):
        """Unpublish the image.

        :param file_name: File name to unpublish.
        """


class LocalPublisher(AbstractPublisher):
    """Image publisher using a local web server."""

    def __init__(self, image_subdir=None, file_permission=0o644,
                 dir_permission=0o755, root_url=None):
        """Create a local publisher.

        :param image_subdir: A subdirectory to put the image to.
            Using an empty directory may cause name conflicts.
        :param file_permission: Permissions for copied files.
        :param dir_permission: Permissions for created directories.
        :param root_url: Public URL of the web server. If empty, determined
            from the configuration.
        """
        self.image_subdir = image_subdir
        self.root_url = (root_url or CONF.deploy.external_http_url
                         or CONF.deploy.http_url)
        self.file_permission = file_permission
        self.dir_permission = dir_permission

    def publish(self, source_path, file_name=None):
        if not file_name:
            file_name = os.path.basename(source_path)

        if self.image_subdir:
            public_dir = os.path.join(CONF.deploy.http_root, self.image_subdir)
        else:
            public_dir = CONF.deploy.http_root

        if not os.path.exists(public_dir):
            os.mkdir(public_dir, self.dir_permission)

        published_file = os.path.join(public_dir, file_name)

        try:
            os.link(source_path, published_file)
            os.chmod(source_path, self.file_permission)
            try:
                utils.execute(
                    '/usr/sbin/restorecon', '-i', '-R', 'v', public_dir)
            except FileNotFoundError as exc:
                LOG.debug(
                    "Could not restore SELinux context on "
                    "%(public_dir)s, restorecon command not found.\n"
                    "Error: %(error)s",
                    {'public_dir': public_dir,
                        'error': exc})

        except OSError as exc:
            LOG.debug(
                "Could not hardlink image file %(image)s to public "
                "location %(public)s (will copy it over): "
                "%(error)s", {'image': source_path,
                              'public': published_file,
                              'error': exc})

            shutil.copyfile(source_path, published_file)
            os.chmod(published_file, self.file_permission)

        if self.image_subdir:
            return os.path.join(self.root_url, self.image_subdir, file_name)
        else:
            return os.path.join(self.root_url, file_name)

    def unpublish(self, file_name):
        published_file = os.path.join(
            CONF.deploy.http_root, self.image_subdir, file_name)
        utils.unlink_without_raise(published_file)


class SwiftPublisher(AbstractPublisher):
    """Image publisher using OpenStack Swift."""

    def __init__(self, container, delete_after):
        """Create a Swift publisher.

        :param container: Swift container to use.
        :param delete_after: Number of seconds after which the link will
            no longer be valid.
        """
        self.container = container
        self.delete_after = delete_after

    def _append_filename_param(self, url, filename):
        """Append 'filename=<file>' parameter to given URL.

        Some BMCs seem to validate boot image URL requiring the URL to end
        with something resembling ISO image file name.

        This function tries to add, hopefully, meaningless 'filename'
        parameter to URL's query string in hope to make the entire boot image
        URL looking more convincing to the BMC.

        However, `url` with fragments might not get cured by this hack.

        :param url: a URL to work on
        :param filename: name of the file to append to the URL
        :returns: original URL with 'filename' parameter appended
        """
        parsed_url = urlparse.urlparse(url)
        parsed_qs = urlparse.parse_qsl(parsed_url.query)

        has_filename = [x for x in parsed_qs if x[0].lower() == 'filename']
        if has_filename:
            return url

        parsed_qs.append(('filename', filename))
        parsed_url = list(parsed_url)
        parsed_url[4] = urlparse.urlencode(parsed_qs)

        return urlparse.urlunparse(parsed_url)

    def publish(self, source_path, file_name=None):
        api = swift.SwiftAPI()
        if not file_name:
            file_name = os.path.basename(source_path)

        object_headers = {'X-Delete-After': str(self.delete_after)}
        api.create_object(self.container, file_name, source_path,
                          object_headers=object_headers)

        image_url = api.get_temp_url(self.container, file_name,
                                     self.delete_after)
        return self._append_filename_param(
            image_url, os.path.basename(source_path))

    def unpublish(self, file_name):
        api = swift.SwiftAPI()
        LOG.debug("Cleaning up image %(name)s from Swift container "
                  "%(container)s", {'name': file_name,
                                    'container': self.container})

        try:
            api.delete_object(self.container, file_name)

        except exception.SwiftOperationError as exc:
            LOG.warning("Failed to clean up image %(image)s. Error: "
                        "%(error)s.", {'image': file_name, 'error': exc})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_config import cfg

from ironic.common.i18n import _


group = cfg.OptGroup(name='oci',
                     title='OCI Container Registry Client Options')
opts = [
    cfg.ListOpt('secure_cdn_registries',
                default=[
                    'registry.redhat.io',
                    'registry.access.redhat.com',
                    'docker.io',
                    'registry-1.docker.io',
                ],
                # NOTE(TheJulia): Not a mutable option because this setting
                # impacts how the OCI client navigates configuration handling
                # for these hosts.
                mutable=False,
                help=_('An option which signals to the OCI Container Registry '
                       'client which remote endpoints are fronted by Content '
                       'Distribution Networks which we may receive redirects '
                       'to in order to download the requested artifacts, '
                       'where the OCI client should go ahead and issue the '
                       'download request with authentication headers before '
                       'being asked by the remote server for user '
                       'authentication.')),
    cfg.StrOpt('authentication_config',
               mutable=True,
               help=_('An option which allows pre-shared authorization keys '
                      'to be utilized by the Ironic service to facilitate '
                      'authentication with remote image registries which '
                      'may require authentication for all interactions. '
                      'Ironic will utilize these credentials to access '
                      'general artifacts, but Ironic will *also* prefer '
                      'user credentials, if supplied, for disk images. '
                      'This file is in the same format utilized in the '
                      'container ecosystem for the same purpose. '
                      'Structured as a JSON document with an ``auths`` '
                      'key, with remote registry domain FQDNs as keys, '
                      'and a nested ``auth`` key within that value which '
                      'holds the actual pre-shared secret. Ironic does '
                      'not cache the contents of this file at launch, '
                      'and the file can be updated as Ironic operates '
                      'in the event pre-shared tokens need to be '
                      'regenerated.')),
]


def register_opts(conf):
    conf.register_group(group)
    conf.register_opts(opts, group='oci')
